import copy
import os
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

import prefect.logging.configuration
import prefect.settings
from pydantic import Field, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

if TYPE_CHECKING:
    pass


class ControlFlowSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="CONTROLFLOW_",
        env_file=(
            ""
            if os.getenv("CONTROLFLOW_TEST_MODE")
            else ("~/.controlflow/.env", ".env")
        ),
        extra="ignore",
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )


class Settings(ControlFlowSettings):
    max_iterations: Optional[int] = Field(
        default=100,
        description="The maximum number of iterations to attempt to run "
        "before raising an error. If None, the system will run indefinitely. ",
    )

    # ------------ home settings ------------

    home_path: Path = Field(
        default="~/.controlflow",
        description="The path to the ControlFlow home directory.",
        validate_default=True,
    )

    # ------------ display and logging settings ------------

    log_prints: bool = Field(
        False,
        description="Whether to log workflow prints to the Prefect logger by default.",
    )

    # ------------ flow settings ------------

    eager_mode: bool = Field(
        default=True,
        description="If True, @task- and @flow-decorated functions are run immediately. "
        "This can be set on a per-task or per-flow basis using the `eager` argument.",
    )
    enable_local_input: bool = Field(
        default=True,
        description="If True, the user can provide input via "
        "the terminal. Otherwise, only API input is accepted.",
    )
    strict_flow_context: bool = Field(
        default=False,
        description="If False, calling Task.run() outside a flow context will automatically "
        "create a flow and run the task within it. If True, an error will be raised.",
    )

    # ------------ LLM settings ------------

    llm_model: str = Field(default="openai/gpt-4o", description="The LLM model to use.")
    llm_temperature: float = Field(0.7, description="The temperature for LLM sampling.")
    max_input_tokens: int = Field(
        100_000, description="The maximum number of tokens to send to an LLM."
    )

    # ------------ Flow visualization settings ------------

    enable_print_handler: bool = Field(
        default=True,
        description="If True, the PrintHandler will be enabled.",
    )

    enable_experimental_tui: bool = Field(
        default=False,
        description="If True, the experimental TUI will be enabled. If False, the TUI will be disabled.",
    )
    run_tui_headless: bool = Field(
        default=False,
        description="If True, the experimental TUI will run in headless mode, which is useful for debugging.",
    )

    # ------------ Debug settings ------------

    tools_raise_on_error: bool = Field(
        False, description="If True, an error in a tool call will raise an exception."
    )

    tools_verbose: bool = Field(
        False, description="If True, tools will log additional information."
    )

    # ------------ Prefect settings ------------
    #
    # Default settings for Prefect when used with ControlFlow. They can be
    # overridden by setting standard Prefect env vars

    prefect_log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = Field(
        default="WARNING",
        description="The log level for Prefect.",
        alias="PREFECT_LOGGING_LEVEL",
    )

    _prefect_context: contextmanager = None

    @field_validator("home_path", mode="before")
    def _validate_home_path(cls, v: Union[str, Path]) -> Path:
        v = Path(v).expanduser()
        if not v.exists():
            v.mkdir(parents=True, exist_ok=True)
        return v

    @model_validator(mode="after")
    def _apply_prefect_settings(self):
        """
        Prefect settings are set at runtime by opening a settings context.
        We check if any prefect-specific settings have been changed and apply them.
        """
        if self._prefect_context is not None:
            self._prefect_context.__exit__(None, None, None)
            self._prefect_context = None

        settings_map = {"prefect_log_level": prefect.settings.PREFECT_LOGGING_LEVEL}

        prefect_settings = {}

        for cf_setting, v in self.model_dump().items():
            if cf_setting.startswith("prefect_"):
                p_setting = settings_map[cf_setting]
                if p_setting.value() != v:
                    prefect_settings[settings_map[cf_setting]] = v

        if prefect_settings:
            self._prefect_context = prefect.settings.temporary_settings(
                prefect_settings
            )
            self._prefect_context.__enter__()

        # Configure logging
        prefect.logging.configuration.setup_logging()

        return self


settings = Settings()


@contextmanager
def temporary_settings(**kwargs: Any):
    """
    Temporarily override ControlFlow setting values.

    Args:
        **kwargs: The settings to override, including nested settings.

    Example:
        Temporarily override a setting:
        ```python
        import controlflow
        from controlflow.settings import temporary_settings

        with temporary_settings(tools_raise_on_error=True):
            assert controlflow.settings.tools_raise_on_error is True
        assert controlflow.settings.tools_raise_on_error is False
        ```
    """
    old_settings = copy.deepcopy(settings.model_dump(exclude={"_prefect_context"}))

    try:
        # apply the new settings
        for attr, value in kwargs.items():
            if not hasattr(settings, attr):
                raise AttributeError(f"Setting {attr} does not exist.")
            setattr(settings, attr, value)
        yield

    finally:
        # restore the old settings
        for attr in kwargs:
            if hasattr(settings, attr):
                setattr(settings, attr, old_settings[attr])



import inspect

from controlflow.agents import Agent
from controlflow.flows import Flow
from controlflow.tasks.task import Task
from controlflow.utilities.jinja import jinja_env
from controlflow.utilities.types import ControlFlowModel

from .controller import Controller


class Template(ControlFlowModel):
    template: str

    def should_render(self) -> bool:
        return True

    def render(self) -> str:
        if self.should_render():
            render_kwargs = dict(self)
            render_kwargs.pop("template")
            return jinja_env.from_string(inspect.cleandoc(self.template)).render(
                **render_kwargs
            )


class AgentTemplate(Template):
    template: str = """
        You are an AI agent participating in a workflow. Your role is to work on
        your tasks and use the provided tools to complete those tasks and
        communicate with the orchestrator. 
        
        Important: The orchestrator is a Python script and cannot read or
        respond to messages posted in this thread. You must use the provided
        tools to communicate with the orchestrator. Posting messages in this
        thread should only be used for thinking out loud, working through a
        problem, or communicating with other agents. Any System messages or
        messages prefixed with "SYSTEM:" are from the workflow system, not an
        actual human.
        
        Your job is to:
        1. Select one or more tasks to work on from the ready tasks.
        2. Read the task instructions and work on completing the task objective, which may
        involve using appropriate tools or collaborating with other agents
        assigned to the same task.
        3. When you (and any other agents) have completed the task objective,
        use the provided tool to inform the orchestrator of the task completion
        and result.
        4. Repeat steps 1-3 until no more tasks are available for execution.        
        
        Note that the orchestrator may decide to activate a different agent at any time.
        
        ## Your information 
        
        - ID: {{ agent.id }}
        - Name: "{{ agent.name }}"
        {% if agent.description -%}
        - Description: "{{ agent.description }}"
        {% endif %}
            
        ## Instructions
        
        You must follow instructions at all times. Instructions can be added or removed at any time.
        
        - Never impersonate another agent 
        
        {% if agent.instructions %}
        {{ agent.instructions }}
        {% endif %}
        
        {% if additional_instructions %}     
        {% for instruction in additional_instructions %}
        - {{ instruction }}
        {% endfor %}
        {% endif %}                  
        """
    agent: Agent
    additional_instructions: list[str]


# class MemoryTemplate(Template):
#     template: str = """
#         ## Memory

#         You have the following private memories:

#         {% for index, memory in memories %}
#         - {{ index }}: {{ memory }}
#         {% endfor %}

#         Use your memory to record information
#     """
#     memories: dict[int, str]


class WorkflowTemplate(Template):
    template: str = """
        
        ## Tasks
        
        As soon as you have completed a task's objective, you must use the provided
        tool to mark it successful and provide a result. It may take multiple
        turns or collaboration with other agents to complete a task. Any agent
        assigned to a task can complete it. Once a task is complete, no other
        agent can interact with it.                 
        
        Tasks should only be marked failed due to technical errors like a broken
        or erroring tool or unresponsive human.
        
        Tasks are not ready until all of their dependencies are met. Parent
        tasks depend on all of their subtasks.
                
        ## Flow
        
        Name: {{ flow.name }}
        {% if flow.description %}
        Description: {{ flow.description }} 
        {% endif %}
        {% if flow.context %}
        Context:
        {% for key, value in flow.context.items() %}
        - {{ key }}: {{ value }}
        {% endfor %}
        {% endif %}
        
        ## Tasks
        
        ### Ready tasks
        
        These tasks are ready to be worked on because all of their dependencies have
        been completed. You can only work on tasks to which you are assigned.
                
        {% for task in ready_tasks %}
        #### Task {{ task.id }}
        - objective: {{ task.objective }}
        - instructions: {{ task.instructions}}
        - context: {{ task.context }}
        - result_type: {{ task.result_type }}
        - depends_on: {{ task.depends_on }}
        - parent: {{ task.parent }}
        - assigned agents: {{ task.agents }}
        {% if task.user_access %} 
        - user access: True 
        {% endif %}
        - created_at: {{ task.created_at }}
        
        {% endfor %}
        
        ### Upstream tasks
        
        {% for task in upstream_tasks %}
        #### Task {{ task.id }}
        - objective: {{ task.objective }}
        - instructions: {{ task.instructions}}
        - status: {{ task.status }}
        - result: {{ task.result }}
        - error: {{ task.error }}
        - context: {{ task.context }}
        - depends_on: {{ task.depends_on }}
        - parent: {{ task.parent }}
        - assigned agents: {{ task.agents }}
        {% if task.user_access %} 
        - user access: True 
        {% endif %}
        - created_at: {{ task.created_at }}
        
        {% endfor %}
        
        ### Downstream tasks
                
        {% for task in downstream_tasks %}
        #### Task {{ task.id }}
        - objective: {{ task.objective }}
        - instructions: {{ task.instructions}}
        - status: {{ task.status }}
        - result_type: {{ task.result_type }}
        - context: {{ task.context }}
        - depends_on: {{ task.depends_on }}
        - parent: {{ task.parent }}
        - assigned agents: {{ task.agents }}
        {% if task.user_access %} 
        - user access: True 
        {% endif %}
        - created_at: {{ task.created_at }}
        
        {% endfor %}
        """

    ready_tasks: list[dict]
    upstream_tasks: list[dict]
    downstream_tasks: list[dict]
    current_task: Task
    flow: Flow


class ToolTemplate(Template):
    template: str = """
        You have access to various tools. They may change, so do not rely on history 
        to see what tools are available.
        
        ## Talking to human users
        
        If your task requires you to interact with a user, it will show
        `user_access=True` and you will be given a `talk_to_user` tool. You can
        use it to send messages to the user and optionally wait for a response.
        This is how you tell the user things and ask questions. Do not mention
        your tasks or the workflow. The user can only see messages you send
        them via tool. They can not read the rest of the
        thread. 

        Human users may give poor, incorrect, or partial responses. You may need
        to ask questions multiple times in order to complete your tasks. Do not
        make up answers for omitted information; ask again and only fail the
        task if you truly can not make progress. If your task requires human
        interaction and neither it nor any assigned agents have `user_access`,
        you can fail the task.
        """

    agent: Agent


class MainTemplate(ControlFlowModel):
    agent: Agent
    controller: Controller
    ready_tasks: list[Task]
    current_task: Task
    context: dict
    instructions: list[str]
    agent_assignments: dict[Task, list[Agent]]

    def render(self):
        # get up to 50 upstream and 50 downstream tasks
        g = self.controller.graph
        upstream_tasks = g.topological_sort([t for t in g.tasks if t.is_complete()])[
            -50:
        ]
        downstream_tasks = g.topological_sort(
            [t for t in g.tasks if t.is_incomplete() and t not in self.ready_tasks]
        )[:50]

        ready_tasks = [t.model_dump() for t in self.ready_tasks]
        upstream_tasks = [t.model_dump() for t in upstream_tasks]
        downstream_tasks = [t.model_dump() for t in downstream_tasks]

        # update agent assignments
        assignments = {t.id: a for t, a in self.agent_assignments.items()}
        for t in ready_tasks + upstream_tasks + downstream_tasks:
            if t["id"] in assignments:
                t["agents"] = assignments[t["id"]]

        templates = [
            AgentTemplate(
                agent=self.agent,
                additional_instructions=self.instructions,
            ),
            WorkflowTemplate(
                flow=self.controller.flow,
                ready_tasks=ready_tasks,
                upstream_tasks=upstream_tasks,
                downstream_tasks=downstream_tasks,
                current_task=self.current_task,
            ),
            ToolTemplate(agent=self.agent),
            # CommunicationTemplate(
            #     agent=self.agent,
            # ),
        ]

        rendered = [
            template.render() for template in templates if template.should_render()
        ]
        return "\n\n".join(rendered)



import controlflow
from controlflow.agents import Agent, get_default_agent
from controlflow.agents.names import NAMES
from controlflow.instructions import instructions
from controlflow.tasks.task import Task
from langchain_openai import ChatOpenAI


class TestAgentInitialization:
    def test_agent_gets_random_name(self):
        agent = Agent()

        assert agent.name in NAMES

    def test_agent_default_model(self):
        agent = Agent()

        # None indicates it will be loaded from the default model
        assert agent.model is None
        assert agent.get_model() is controlflow.get_default_model()

    def test_agent_model(self):
        model = ChatOpenAI(model="gpt-3.5-turbo")
        agent = Agent(model=model)

        # None indicates it will be loaded from the default model
        assert agent.model is model
        assert agent.get_model() is model

    def test_agent_loads_instructions_at_creation(self):
        with instructions("test instruction"):
            agent = Agent()

        assert "test instruction" in agent.instructions


class TestDefaultAgent:
    def test_default_agent(self):
        assert get_default_agent().name == "Marvin"
        assert Task("task").get_agents()[0] is get_default_agent()

    def test_default_agent_has_no_tools(self):
        assert get_default_agent().tools == []

    def test_default_agent_has_no_model(self):
        assert get_default_agent().model is None

    def test_default_agent_can_be_assigned(self):
        # baseline
        assert get_default_agent().name == "Marvin"

        new_default_agent = Agent(name="New Agent")
        controlflow.default_agent = new_default_agent

        assert get_default_agent().name == "New Agent"
        assert Task("task").get_agents()[0] is new_default_agent
        assert [a.name for a in Task("task").get_agents()] == ["New Agent"]

    def test_updating_the_default_model_updates_the_default_agent_model(self):
        new_model = ChatOpenAI(model="gpt-3.5-turbo")
        controlflow.default_model = new_model

        new_agent = get_default_agent()
        assert new_agent.model is None
        assert new_agent.get_model() is new_model

        task = Task("task")
        assert task.get_agents()[0].model is None
        assert task.get_agents()[0].get_model() is new_model



import datetime

import rich
from rich import box
from rich.console import Group
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.spinner import Spinner
from rich.table import Table

import controlflow
from controlflow.llm.handlers import CompletionHandler
from controlflow.llm.messages import (
    AIMessage,
    AIMessageChunk,
    MessageType,
    SystemMessage,
    ToolCall,
    ToolMessage,
    UserMessage,
)
from controlflow.utilities.rich import console as cf_console


class PrintHandler(CompletionHandler):
    def __init__(self):
        self.messages: dict[str, MessageType] = {}
        self.live: Live = Live(auto_refresh=True, console=cf_console)
        self.paused_id: str = None
        super().__init__()

    def on_start(self):
        try:
            self.live.start()
        except rich.errors.LiveError:
            pass

    def on_end(self):
        self.live.stop()

    def on_exception(self, exc: Exception):
        self.live.stop()

    def update_live(self, latest: MessageType = None):
        # sort by timestamp, using the custom message id as a tiebreaker
        # in case the same message appears twice (e.g. tool call and message)
        messages = sorted(self.messages.items(), key=lambda m: (m[1].timestamp, m[0]))
        content = []

        tool_results = {}  # To track tool results by their call ID

        # gather all tool messages first
        for _, message in messages:
            if isinstance(message, ToolMessage):
                tool_results[message.tool_call_id] = message

        for _, message in messages:
            if isinstance(message, (SystemMessage, UserMessage, AIMessage)):
                content.append(format_message(message, tool_results=tool_results))
            # no need to handle tool messages

        if self.live.is_started:
            self.live.update(Group(*content), refresh=True)
        elif latest:
            cf_console.print(format_message(latest))

    def on_message_delta(self, delta: AIMessageChunk, snapshot: AIMessageChunk):
        self.messages[snapshot.id] = snapshot
        self.update_live()

    def on_message_done(self, message: AIMessage):
        self.messages[message.id] = message
        self.update_live(latest=message)

    def on_tool_call_delta(self, delta: AIMessageChunk, snapshot: AIMessageChunk):
        self.messages[snapshot.id] = snapshot
        self.update_live()

    def on_tool_call_done(self, message: AIMessage):
        self.messages[message.id] = message
        self.update_live(latest=message)

    def on_tool_result_created(self, message: AIMessage, tool_call: ToolCall):
        # if collecting input on the terminal, pause the live display
        # to avoid overwriting the input prompt
        if tool_call["name"] == "talk_to_user":
            self.paused_id = tool_call["id"]
            self.live.stop()
            self.messages.clear()

    def on_tool_result_done(self, message: ToolMessage):
        self.messages[f"tool-result:{message.tool_call_id}"] = message

        # if we were paused, resume the live display
        if self.paused_id and self.paused_id == message.tool_call_id:
            self.paused_id = None
            # print newline to avoid odd formatting issues
            print()
            self.live = Live(auto_refresh=False)
            self.live.start()
        self.update_live(latest=message)


def format_timestamp(timestamp: datetime.datetime) -> str:
    local_timestamp = timestamp.astimezone()
    return local_timestamp.strftime("%I:%M:%S %p").lstrip("0").rjust(11)


def status(icon, text) -> Table:
    t = Table.grid(padding=1)
    t.add_row(icon, text)
    return t


ROLE_COLORS = {
    "system": "gray",
    "ai": "blue",
    "user": "green",
}
ROLE_NAMES = {
    "system": "System",
    "ai": "Agent",
    "user": "User",
}


def format_message(message: MessageType, tool_results: dict = None) -> Panel:
    if message.role == "ai" and message.name:
        title = f"Agent: {message.name}"
    else:
        title = ROLE_NAMES.get(message.role, "Agent")

    content = []
    if message.str_content:
        content.append(Markdown(message.str_content or ""))

    tool_content = []
    for tool_call in getattr(message, "tool_calls", []):
        tool_result = (tool_results or {}).get(tool_call["id"])
        if tool_result:
            c = format_tool_result(tool_result)

        else:
            c = format_tool_call(tool_call)
        if c:
            tool_content.append(c)

    if content and tool_content:
        content.append("\n")

    return Panel(
        Group(*content, *tool_content),
        title=f"[bold]{title}[/]",
        subtitle=f"[italic]{format_timestamp(message.timestamp)}[/]",
        title_align="left",
        subtitle_align="right",
        border_style=ROLE_COLORS.get(message.role, "red"),
        box=box.ROUNDED,
        width=100,
        expand=True,
        padding=(1, 2),
    )


def format_tool_call(tool_call: ToolCall) -> Panel:
    name = tool_call["name"]
    args = tool_call["args"]
    if controlflow.settings.tools_verbose:
        return status(Spinner("dots"), f'Tool call: "{name}"\n\nTool args: {args}')
    return status(Spinner("dots"), f'Tool call: "{name}"')


def format_tool_result(message: ToolMessage) -> Panel:
    name = message.tool_call["name"]

    if message.is_error:
        icon = ":x:"
    else:
        icon = ":white_check_mark:"

    if controlflow.settings.tools_verbose:
        msg = f'Tool call: "{name}"\n\nTool result: {message.str_content}'
    else:
        msg = f'Tool call: "{name}"'
    return status(icon, msg)



import datetime

import controlflow
from controlflow.llm.messages import AIMessage, ToolMessage
from controlflow.utilities.testing import record_messages


def test_record_messages_empty():
    with record_messages() as messages:
        pass
    assert messages == []


def test_record_task_messages(default_fake_llm):
    task = controlflow.Task("say hello")

    response = AIMessage(
        agent=dict(name="Marvin"),
        id="run-2af8bb73-661f-4ec3-92ff-d7d8e3074926",
        timestamp=datetime.datetime(
            2024, 6, 23, 17, 12, 24, 91830, tzinfo=datetime.timezone.utc
        ),
        role="ai",
        content="",
        name="Marvin",
        tool_calls=[
            {
                "name": f"mark_task_{task.id}_successful",
                "args": {"result": "Hello!"},
                "id": "call_ZEPdV8mCgeBe5UHjKzm6e3pe",
            }
        ],
        is_delta=False,
    )

    default_fake_llm.set_responses([response])
    with record_messages() as rec_messages:
        task.run()

    assert rec_messages[0].content == response.content
    assert rec_messages[0].id == response.id
    assert rec_messages[0].tool_calls == response.tool_calls

    expected_tool_message = ToolMessage(
        agent=dict(name="Marvin"),
        id="cb84bb8f3e0f4245bbf5eefeee9272b2",
        timestamp=datetime.datetime(
            2024, 6, 23, 17, 12, 24, 187384, tzinfo=datetime.timezone.utc
        ),
        role="tool",
        content=f'Task {task.id} ("say hello") marked successful by Marvin.',
        name="Marvin",
        tool_call_id="call_ZEPdV8mCgeBe5UHjKzm6e3pe",
        tool_call={
            "name": f"mark_task_{task.id}_successful",
            "args": {"result": "Hello!"},
            "id": "call_ZEPdV8mCgeBe5UHjKzm6e3pe",
        },
        tool_metadata={"ignore_result": True},
    )
    assert rec_messages[1].content == expected_tool_message.content
    assert rec_messages[1].tool_call_id == expected_tool_message.tool_call_id
    assert rec_messages[1].tool_call == expected_tool_message.tool_call



import pytest
from controlflow.llm.messages import MessageType
from controlflow.settings import temporary_settings
from prefect.testing.utilities import prefect_test_harness

from .fixtures import *


@pytest.fixture(autouse=True, scope="session")
def temp_controlflow_settings():
    with temporary_settings(max_iterations=10, enable_experimental_tui=False):
        yield


@pytest.fixture(autouse=True)
def reset_settings_after_each_test():
    with temporary_settings():
        yield


@pytest.fixture(autouse=True, scope="session")
def prefect_test_fixture():
    """
    Run Prefect against temporary sqlite database
    """
    with prefect_test_harness():
        yield



from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
)
from uuid import UUID

import prefect
import prefect.cache_policies
import prefect.serializers
import prefect.tasks
from prefect import get_client as get_prefect_client
from prefect.artifacts import ArtifactRequest
from prefect.client.orchestration import SyncPrefectClient, get_client
from prefect.client.schemas import State, TaskRun
from prefect.context import (
    FlowRunContext,
    TaskRunContext,
)
from prefect.events.schemas.events import Event
from prefect.states import (
    Cancelled,
    Completed,
    Failed,
    Running,
)
from prefect.utilities.asyncutils import run_coro_as_sync
from prefect.utilities.engine import (
    emit_task_run_state_change_event,
    propose_state_sync,
)
from pydantic import TypeAdapter

import controlflow
from controlflow.utilities.types import ControlFlowModel

if TYPE_CHECKING:
    pass


def prefect_task(*args, **kwargs):
    """
    A decorator that creates a Prefect task with ControlFlow defaults
    """

    # TODO: only open in Flow context?

    kwargs.setdefault("log_prints", controlflow.settings.log_prints)
    kwargs.setdefault("cache_policy", prefect.cache_policies.NONE)
    kwargs.setdefault("result_serializer", prefect.serializers.JSONSerializer())

    return prefect.task(*args, **kwargs)


def prefect_flow(*args, **kwargs):
    """
    A decorator that creates a Prefect flow with ControlFlow defaults
    """

    kwargs.setdefault("log_prints", controlflow.settings.log_prints)

    return prefect.flow(*args, **kwargs)


def create_markdown_artifact(
    key: str,
    markdown: str,
    description: str = None,
    task_run_id: UUID = None,
    flow_run_id: UUID = None,
) -> None:
    """
    Create a Markdown artifact.
    """
    tr_context = TaskRunContext.get()
    fr_context = FlowRunContext.get()

    if tr_context:
        task_run_id = task_run_id or tr_context.task_run.id
    if fr_context:
        flow_run_id = flow_run_id or fr_context.flow_run.id

    client = get_prefect_client()
    run_coro_as_sync(
        client.create_artifact(
            artifact=ArtifactRequest(
                key=key,
                data=markdown,
                description=description,
                type="markdown",
                task_run_id=task_run_id,
                flow_run_id=flow_run_id,
            )
        )
    )


def create_json_artifact(
    key: str,
    data: Any,
    description: str = None,
    task_run_id: UUID = None,
    flow_run_id: UUID = None,
) -> None:
    """
    Create a JSON artifact.
    """

    try:
        markdown = TypeAdapter(type(data)).dump_json(data, indent=2).decode()
        markdown = f"```json\n{markdown}\n```"
    except Exception:
        markdown = str(data)

    create_markdown_artifact(
        key=key,
        markdown=markdown,
        description=description,
        task_run_id=task_run_id,
        flow_run_id=flow_run_id,
    )


def create_python_artifact(
    key: str,
    code: str,
    description: str = None,
    task_run_id: UUID = None,
    flow_run_id: UUID = None,
) -> None:
    """
    Create a Python artifact.
    """

    create_markdown_artifact(
        key=key,
        markdown=f"```python\n{code}\n```",
        description=description,
        task_run_id=task_run_id,
        flow_run_id=flow_run_id,
    )


class PrefectTrackingTask(ControlFlowModel):
    """
    A utility for creating a Prefect task that tracks the state of a task run
    without requiring a function or related Prefect machinery such as error
    handling. We use this to map ControlFlow Task objects onto Prefect states.

    While calling cf.Task.run() provides an opportunity to map a ControlFlow
    task directly onto an invocation of compute (and therefore a classic Prefect
    Task), when run as part of a flow we do not "invoke" ControlFlow tasks
    directly. Instead, an agent may decide at any time to work on any ready task
    and, ultimately, mark it as complete. We model this behavior into Prefect
    with this TrackingTask. The corresponding Prefect task starts "running" as
    soon as the CF Task is ready to run (e.g. as soon as it becomes eligible for
    an agent to work on it). It then transitions to a terminal state when the CF
    Task is complete. This gives us excellent visibility into the state of the
    CF Task within the Prefect UI, even though it doesn't correspond to a single
    invocation of compute.
    """

    name: str
    description: Optional[str] = None
    task_run_id: Optional[str] = None
    tags: Optional[list[str]] = None

    _task: prefect.Task = None
    _task_run: TaskRun = None
    _client: SyncPrefectClient = None
    _last_event: Optional[Event] = None
    is_started: bool = False

    _context: list = []

    def start(self, depends_on: list = None):
        if self.is_started:
            raise ValueError("Task already started")
        self.is_started = True
        self._client = get_client(sync_client=True)

        self._task = prefect.Task(
            fn=lambda: None,
            name=self.name,
            description=self.description,
            tags=self.tags,
        )

        self._task_run = run_coro_as_sync(
            self._task.create_run(
                id=self.task_run_id,
                parameters=dict(depends_on=depends_on),
                flow_run_context=FlowRunContext.get(),
            )
        )

        self._last_event = emit_task_run_state_change_event(
            task_run=self._task_run,
            initial_state=None,
            validated_state=self._task_run.state,
        )

        self.set_state(Running())

    def set_state(self, state: State) -> State:
        if not self.is_started:
            raise ValueError("Task not started")

        new_state = propose_state_sync(
            self._client, state, task_run_id=self._task_run.id, force=True
        )

        self._last_event = emit_task_run_state_change_event(
            task_run=self._task_run,
            initial_state=self._task_run.state,
            validated_state=new_state,
            follows=self._last_event,
        )
        self._task_run.state = new_state
        return new_state

    def succeed(self, result: Any):
        terminal_state = Completed()
        self.set_state(terminal_state)

    def fail(self, error: Optional[str] = None):
        self.set_state(Failed(message=error))

    def skip(self):
        self.set_state(Cancelled(message="Task skipped"))


def prefect_task_context(**kwargs):
    """
    Creates a Prefect task that starts when the context is entered and ends when
    it closes. This is useful for creating a Prefect task that is not tied to a
    specific function but governs a block of code. Note that some features, like
    retries and caching, will not work.
    """
    supported_kwargs = {
        "name",
        "description",
        "task_run_name",
        "tags",
        "version",
        "timeout_seconds",
        "log_prints",
        "on_completion",
        "on_failure",
    }
    unsupported_kwargs = set(kwargs.keys()) - set(supported_kwargs)
    if unsupported_kwargs:
        raise ValueError(
            f"Unsupported keyword arguments for a task context provided: "
            f"{unsupported_kwargs}. Consider using a @task-decorated function instead."
        )

    @contextmanager
    @prefect_task(**kwargs)
    def task_context():
        yield

    return task_context()


def prefect_flow_context(**kwargs):
    """
    Creates a Prefect flow that starts when the context is entered and ends when
    it closes. This is useful for creating a Prefect flow that is not tied to a
    specific function but governs a block of code. Note that some features, like
    retries and caching, will not work.
    """

    supported_kwargs = {
        "name",
        "description",
        "flow_run_name",
        "tags",
        "version",
        "timeout_seconds",
        "log_prints",
        "on_completion",
        "on_failure",
    }
    unsupported_kwargs = set(kwargs.keys()) - set(supported_kwargs)
    if unsupported_kwargs:
        raise ValueError(
            f"Unsupported keyword arguments for a flow context provided: "
            f"{unsupported_kwargs}. Consider using a @flow-decorated function instead."
        )

    @contextmanager
    @prefect_flow(**kwargs)
    def flow_context():
        yield

    return flow_context()



# --- Public top-level API ---

from .settings import settings

from .agents import Agent
from .tasks import Task
from .flows import Flow

from .instructions import instructions
from .decorators import flow, task
from .llm.tools import tool

# --- Default settings ---

from .llm.models import _get_initial_default_model, get_default_model
from .flows.history import InMemoryHistory, get_default_history

# assign to controlflow.default_model to change the default model
default_model = _get_initial_default_model()
del _get_initial_default_model

# assign to controlflow.default_history to change the default history
default_history = InMemoryHistory()
del InMemoryHistory

# assign to controlflow.default_agent to change the default agent
default_agent = Agent(name="Marvin")

# --- Version ---

try:
    from ._version import version as __version__  # type: ignore
except ImportError:
    __version__ = "unknown"



from typing import Union

import tiktoken
from langchain_openai import AzureChatOpenAI, ChatOpenAI
from pydantic import TypeAdapter

import controlflow
from controlflow.llm.messages import AIMessage, SystemMessage, UserMessage
from controlflow.llm.models import BaseChatModel


def classify(
    data: str,
    labels: list,
    instructions: str = None,
    context: dict = None,
    model: BaseChatModel = None,
):
    try:
        label_strings = [TypeAdapter(type(t)).dump_json(t).decode() for t in labels]
    except Exception as exc:
        raise ValueError(f"Unable to cast labels to strings: {exc}")

    messages = [
        SystemMessage(
            """ 
            You are an expert classifier that always maintains as much semantic meaning
            as possible when labeling information. You use inference or deduction whenever
            necessary to understand missing or omitted data. Classify the provided data,
            text, or information as one of the provided labels. For boolean labels,
            consider "truthy" or affirmative inputs to be "true".

            ## Labels
            
            You must classify the data as one of the following labels, which are
            numbered (starting from 0) and provide a brief description. Output
            the label number only.
            
            {% for label in labels %}
            - Label {{ loop.index0 }}: {{ label }}
            {% endfor %}
            """
        ).render(labels=label_strings),
        UserMessage(
            """
            ## Information to classify
            
            {{ data }}
            
            {% if instructions -%}
            ## Additional instructions
            
            {{ instructions }}
            {% endif %}
            
            {% if context -%}
            ## Additional context
            
            {% for key, value in context.items() -%}
            - {{ key }}: {{ value }}
            
            {% endfor %}
            {% endif %}
            
            """
        ).render(data=data, instructions=instructions, context=context),
        AIMessage("""
            The best label for the data is Label number
            """),
    ]

    model = model or controlflow.llm.models.get_default_model()

    kwargs = {}
    if isinstance(model, (ChatOpenAI, AzureChatOpenAI)):
        openai_kwargs = _openai_kwargs(model=model, n_labels=len(labels))
        kwargs.update(openai_kwargs)
    else:
        messages.append(
            SystemMessage(
                "Return only the label number, no other information or tokens."
            )
        )

    result = controlflow.llm.completions.completion(
        messages=messages,
        model=model,
        max_tokens=1,
        **kwargs,
    )

    index = int(result[0].content)
    return labels[index]


def _openai_kwargs(model: Union[AzureChatOpenAI, ChatOpenAI], n_labels: int):
    encoding = tiktoken.encoding_for_model(model.model_name)

    logit_bias = {}
    for i in range(n_labels):
        for token in encoding.encode(str(i)):
            logit_bias[token] = 100

    return dict(logit_bias=logit_bias)



from controlflow import Agent, Task, flow
from controlflow.instructions import instructions

teacher = Agent(name="teacher")
student = Agent(name="student")


@flow
def demo():
    with Task("Teach a class by asking and answering 3 questions") as task:
        for _ in range(3):
            question = Task(
                "ask the student a question. Wait for the student to answer your question before asking another one.",
                str,
                agents=[teacher],
            )
            with instructions("one sentence max"):
                Task(
                    "answer the question",
                    str,
                    agents=[student],
                    context=dict(question=question),
                )

    task.run()
    return task


t = demo()



import logging
from functools import lru_cache
from typing import Optional


@lru_cache()
def get_logger(name: Optional[str] = None) -> logging.Logger:
    """
    Retrieves a logger with the given name, or the root logger if no name is given.

    Args:
        name: The name of the logger to retrieve.

    Returns:
        The logger with the given name, or the root logger if no name is given.

    Example:
        Basic Usage of `get_logger`
        ```python
        from controlflow.utilities.logging import get_logger

        logger = get_logger("controlflow.test")
        logger.info("This is a test") # Output: controlflow.test: This is a test

        debug_logger = get_logger("controlflow.debug")
        debug_logger.debug_kv("TITLE", "log message", "green")
        ```
    """
    parent_logger = logging.getLogger("controlflow")

    if name:
        # Append the name if given but allow explicit full names e.g. "controlflow.test"
        # should not become "controlflow.controlflow.test"
        if not name.startswith(parent_logger.name + "."):
            logger = parent_logger.getChild(name)
        else:
            logger = logging.getLogger(name)
    else:
        logger = parent_logger

    return logger



from typing import Optional, Union

from controlflow.agents.agent import Agent
from controlflow.llm.messages import (
    AIMessage,
    MessageType,
    SystemMessage,
    UserMessage,
)
from controlflow.llm.rules import LLMRules
from controlflow.tools import Tool


def create_system_message(
    content: str, rules: LLMRules
) -> Union[SystemMessage, UserMessage]:
    """
    Creates a SystemMessage or HumanMessage with SYSTEM: prefix, depending on the rules.
    """
    if rules.system_message_must_be_first:
        return SystemMessage(content=content)
    else:
        return UserMessage(content=f"SYSTEM: {content}")


def handle_agent_info_in_messages(
    messages: list[MessageType], agent: Agent, rules: LLMRules
) -> list[MessageType]:
    """
    If the message is from an agent, add a system message immediately before it to clarify which agent
    it is from. This helps the system follow multi-agent conversations.

    """
    if not rules.add_system_messages_for_multi_agent:
        return messages

    current_agent = agent
    new_messages = []
    for msg in messages:
        # if the message is from a different agent than the previous message,
        # add a clarifying system message
        if isinstance(msg, AIMessage) and msg.agent and msg.agent != current_agent:
            system_msg = SystemMessage(
                content=f'The following message is from agent "{msg.agent["name"]}" '
                f'with id {msg.agent["id"]}.'
            )
            new_messages.append(system_msg)
            current_agent = msg.agent
        new_messages.append(msg)
    return new_messages


def handle_system_messages_must_be_first(messages: list[MessageType], rules: LLMRules):
    if rules.system_message_must_be_first:
        new_messages = []
        # consolidate consecutive SystemMessages into one
        if isinstance(messages[0], SystemMessage):
            content = [messages[0].content]
            i = 1
            while i < len(messages) and isinstance(messages[i], SystemMessage):
                i += 1
                content.append(messages[i].content)
            new_messages.append(SystemMessage(content="\n\n".join(content)))

        # replace all other SystemMessages with HumanMessages
        for i, msg in enumerate(messages[len(new_messages) :]):
            if isinstance(msg, SystemMessage):
                msg = UserMessage(content=f"SYSTEM: {msg.content}")
            new_messages.append(msg)

        return new_messages
    else:
        return messages


def handle_user_message_must_be_first_after_system(
    messages: list[MessageType], rules: LLMRules
):
    if rules.user_message_must_be_first_after_system:
        if not messages:
            messages.append(UserMessage(content="SYSTEM: Begin."))

        # else get first non-system message
        else:
            i = 0
            while i < len(messages) and isinstance(messages[i], SystemMessage):
                i += 1
            if i == len(messages) or (
                i < len(messages) and not isinstance(messages[i], UserMessage)
            ):
                messages.insert(i, UserMessage(content="SYSTEM: Begin."))
    return messages


def prepare_messages(
    agent: Agent,
    messages: list[MessageType],
    system_message: Optional[SystemMessage],
    rules: LLMRules,
    tools: list[Tool],
):
    """This is the main function for processing messages. It applies all the rules"""
    messages = messages.copy()

    if system_message is not None:
        messages.insert(0, system_message)

    messages = handle_agent_info_in_messages(messages, agent=agent, rules=rules)

    if not rules.allow_last_message_has_ai_role_with_tools:
        if messages and tools and isinstance(messages[-1], AIMessage):
            messages.append(create_system_message("Continue.", rules=rules))

    if not rules.allow_consecutive_ai_messages:
        if messages:
            i = 1
            while i < len(messages):
                if isinstance(messages[i], AIMessage) and isinstance(
                    messages[i - 1], AIMessage
                ):
                    messages.insert(i, create_system_message("Continue.", rules=rules))
                i += 1

    messages = handle_system_messages_must_be_first(messages, rules=rules)
    messages = handle_user_message_must_be_first_after_system(messages, rules=rules)

    return messages



from controlflow.agents import Agent
from controlflow.flows import Flow, get_flow
from controlflow.llm.messages import UserMessage
from controlflow.tasks.task import Task
from controlflow.utilities.context import ctx


class TestFlowInitialization:
    def test_flow_initialization(self):
        flow = Flow()
        assert flow.thread_id is not None
        assert len(flow.tools) == 0
        assert len(flow.agents) == 0
        assert len(flow.context) == 0

    def test_flow_with_custom_tools(self):
        def tool1():
            pass

        def tool2():
            pass

        flow = Flow(tools=[tool1, tool2])
        assert len(flow.tools) == 2
        assert tool1 in flow.tools
        assert tool2 in flow.tools

    def test_flow_with_custom_context(self):
        flow = Flow(context={"key": "value"})
        assert len(flow.context) == 1
        assert flow.context["key"] == "value"


class TestFlowContext:
    def test_flow_context_manager(self):
        with Flow() as flow:
            assert ctx.get("flow") == flow
            assert ctx.get("tasks") == []
        assert ctx.get("flow") is None
        assert ctx.get("tasks") == []

    def test_get_flow_within_context(self):
        with Flow() as flow:
            assert get_flow() == flow

    def test_get_flow_without_context(self):
        assert get_flow() is None

    def test_reentrant_flow_context(self):
        flow = Flow()
        with flow:
            assert get_flow() is flow
            with flow:
                assert get_flow() is flow
                with flow:
                    assert get_flow() is flow
                assert get_flow() is flow
            assert get_flow() is flow
        assert get_flow() is None

    def test_get_flow_nested_contexts(self):
        with Flow() as flow1:
            assert get_flow() == flow1
            with Flow() as flow2:
                assert get_flow() == flow2
            assert get_flow() == flow1
        assert get_flow() is None

    def test_tasks_created_in_flow_context(self):
        with Flow() as flow:
            t1 = Task("test 1")
            t2 = Task("test 2")

        assert flow.tasks == {t1.id: t1, t2.id: t2}

    def test_tasks_created_in_nested_flows_only_in_inner_flow(self):
        with Flow() as flow1:
            t1 = Task("test 1")
            with Flow() as flow2:
                t2 = Task("test 2")

        assert flow1.tasks == {t1.id: t1}
        assert flow2.tasks == {t2.id: t2}


class TestFlowHistory:
    def test_get_messages_empty(self):
        flow = Flow()
        messages = flow.get_messages()
        assert messages == []

    def test_add_messages_with_history(self):
        flow = Flow()
        flow.add_messages(
            messages=[UserMessage(content="hello"), UserMessage(content="world")]
        )
        messages = flow.get_messages()
        assert len(messages) == 2
        assert [m.content for m in messages] == ["hello", "world"]

    def test_copy_parent_history(self):
        flow1 = Flow()
        flow1.add_messages(
            messages=[UserMessage(content="hello"), UserMessage(content="world")]
        )

        with flow1:
            flow2 = Flow()

        messages1 = flow1.get_messages()
        assert len(messages1) == 2
        assert [m.content for m in messages1] == ["hello", "world"]

        messages2 = flow2.get_messages()
        assert len(messages2) == 2
        assert [m.content for m in messages2] == ["hello", "world"]

    def test_disable_copying_parent_history(self):
        flow1 = Flow()
        flow1.add_messages(
            messages=[UserMessage(content="hello"), UserMessage(content="world")]
        )

        with flow1:
            flow2 = Flow(copy_parent=False)

        messages1 = flow1.get_messages()
        assert len(messages1) == 2
        assert [m.content for m in messages1] == ["hello", "world"]

        messages2 = flow2.get_messages()
        assert len(messages2) == 0

    def test_child_flow_messages_dont_go_to_parent(self):
        flow1 = Flow()
        flow1.add_messages(
            messages=[UserMessage(content="hello"), UserMessage(content="world")]
        )

        with flow1:
            flow2 = Flow()
            flow2.add_messages(messages=[UserMessage(content="goodbye")])

        messages1 = flow1.get_messages()
        assert len(messages1) == 2
        assert [m.content for m in messages1] == ["hello", "world"]

        messages2 = flow2.get_messages()
        assert len(messages2) == 3
        assert [m.content for m in messages2] == ["hello", "world", "goodbye"]


class TestFlowCreatesDefaults:
    def test_flow_with_custom_agents(self):
        agent1 = Agent(name="Agent 1")
        agent2 = Agent(name="Agent 2")
        flow = Flow(agents=[agent1, agent2])
        assert len(flow.agents) == 2
        assert agent1 in flow.agents
        assert agent2 in flow.agents

    def test_flow_agent_becomes_task_default(self):
        agent = Agent()
        t1 = Task("t1")
        assert t1.agents != [agent]

        with Flow(agents=[agent]):
            t2 = Task("t2")
            assert t2.get_agents() == [agent]



import inspect
import logging
import math
from collections import defaultdict
from contextlib import asynccontextmanager
from typing import Callable

from pydantic import Field, PrivateAttr, field_validator, model_validator

import controlflow
from controlflow.agents import Agent
from controlflow.controllers.graph import Graph
from controlflow.controllers.process_messages import prepare_messages
from controlflow.flows import Flow, get_flow
from controlflow.handlers.print_handler import PrintHandler
from controlflow.instructions import get_instructions
from controlflow.llm.completions import completion, completion_async
from controlflow.llm.handlers import ResponseHandler, TUIHandler
from controlflow.llm.messages import MessageType, SystemMessage
from controlflow.tasks.task import Task
from controlflow.tools import as_tools
from controlflow.utilities.context import ctx
from controlflow.utilities.prefect import create_markdown_artifact
from controlflow.utilities.prefect import prefect_task as prefect_task
from controlflow.utilities.types import ControlFlowModel

logger = logging.getLogger(__name__)


def create_messages_markdown_artifact(messages, thread_id):
    markdown_messages = "\n\n".join([f"{msg.role}: {msg.content}" for msg in messages])
    create_markdown_artifact(
        key="messages",
        markdown=inspect.cleandoc(
            """
            # Messages
            
            *Thread ID: {thread_id}*
            
            {markdown_messages}
            """.format(
                thread_id=thread_id,
                markdown_messages=markdown_messages,
            )
        ),
    )


class Controller(ControlFlowModel):
    """
    A controller contains logic for executing agents with context about the
    larger workflow, including the flow itself, any tasks, and any other agents
    they are collaborating with. The controller is responsible for orchestrating
    agent behavior by generating instructions and tools for each agent. Note
    that while the controller accepts details about (potentially multiple)
    agents and tasks, it's responsiblity is to invoke one agent one time. Other
    mechanisms should be used to orchestrate multiple agents invocations. This
    is done by the controller to avoid tying e.g. agents to tasks or even a
    specific flow.
    """

    # the flow is tracked by the Controller, not the Task, so that tasks can be
    # defined and even instantiated outside a flow. When a Controller is
    # created, we know we're inside a flow context and ready to load defaults
    # and run.
    flow: Flow = Field(
        default_factory=get_flow,
        description="The flow that the controller is a part of.",
        validate_default=True,
    )
    tasks: list[Task] = Field(
        description="Tasks that the controller will complete.",
    )
    agents: dict[Task, list[Agent]] = Field(
        default_factory=dict,
        description="Optionally assign agents to complete tasks. The provided mapping must be task"
        " -> [agents]. Any tasks that aren't included will use their default agents.",
    )
    context: dict = {}
    model_config: dict = dict(extra="forbid")
    enable_experimental_tui: bool = Field(
        default_factory=lambda: controlflow.settings.enable_experimental_tui
    )
    max_iterations: int = Field(
        default_factory=lambda: controlflow.settings.max_iterations
    )
    _iteration: int = 0
    _should_stop: bool = False
    _end_turn_counts: dict = PrivateAttr(default_factory=lambda: defaultdict(int))

    @property
    def graph(self) -> Graph:
        return Graph.from_tasks(self.flow.tasks.values())

    @field_validator("agents", mode="before")
    def _default_agents(cls, v):
        if v is None:
            v = {}
        return v

    @model_validator(mode="after")
    def _finalize(self):
        for task in self.tasks:
            self.flow.add_task(task)
        return self

    def _create_end_turn_tool(self) -> Callable:
        def end_turn():
            """
            This tool is for emergencies only; you should not use it normally.
            If you find yourself in a situation where you are repeatedly invoked
            and your normal tools do not work, or you can not escape the loop,
            use this tool to signal to the controller that you are stuck. A new
            agent will be selected to go next. If this tool is used 3 times by
            an agent the workflow will be aborted automatically.

            """

            # the agent's name is used as the key to track the number of times
            key = getattr(ctx.get("agent", None), "name", None)

            self._end_turn_counts[key] += 1
            if self._end_turn_counts[key] >= 3:
                self._should_stop = True
                self._end_turn_counts[key] = 0

            return (
                f"Ending turn. {3 - self._end_turn_counts[key]}"
                " more uses will abort the workflow."
            )

        return end_turn

    @asynccontextmanager
    async def tui(self):
        if tui := ctx.get("tui"):
            yield tui
        elif self.enable_experimental_tui:
            from controlflow.tui.app import TUIApp as TUI

            tui = TUI(flow=self.flow)
            with ctx(tui=tui):
                async with tui.run_context():
                    yield tui
        else:
            yield

    def _setup_run(self):
        """
        Generate the payload for a single run of the controller.
        """
        if self._iteration >= (self.max_iterations or math.inf):
            raise ValueError(
                f"Controller has exceeded maximum iterations of {self.max_iterations}."
            )
        ready_tasks = [t for t in self.tasks if t.is_ready()]

        # if there are no ready tasks, return. This will usually happen because
        # all the tasks are complete.
        if not ready_tasks:
            return

        # start tracking tasks
        for task in ready_tasks:
            if not task._prefect_task.is_started:
                task._prefect_task.start(
                    depends_on=[
                        t.result for t in task.depends_on if t.result is not None
                    ]
                )

        messages = self.flow.get_messages()

        # get an agent from the next ready task
        agents = self.agents.get(ready_tasks[0], None)
        if agents is None:
            agents = ready_tasks[0].get_agents()
        if len(agents) == 1:
            agent = agents[0]
        else:
            strategy_fn = ready_tasks[0].get_agent_strategy()
            agent = strategy_fn(agents=agents, task=ready_tasks[0], flow=self.flow)
            ready_tasks[0]._iteration += 1

        from controlflow.controllers.instruction_template import MainTemplate

        tools = self.flow.tools + agent.get_tools() + [self._create_end_turn_tool()]

        # add tools for any ready tasks that the agent is assigned to
        for task in ready_tasks:
            if agent in self.agents.get(task, []) or agent in task.get_agents():
                tools.extend(task.get_tools())

        instructions_template = MainTemplate(
            agent=agent,
            controller=self,
            ready_tasks=ready_tasks,
            current_task=ready_tasks[0],
            context=self.context,
            instructions=get_instructions(),
            agent_assignments=self.agents,
        )
        instructions = instructions_template.render()

        # prepare messages
        system_message = SystemMessage(content=instructions)

        rules = agent.get_llm_rules()

        messages = prepare_messages(
            agent=agent,
            system_message=system_message,
            messages=messages,
            rules=rules,
            tools=tools,
        )

        # setup handlers
        handlers = []
        if self.enable_experimental_tui:
            handlers.append(TUIHandler())
        elif controlflow.settings.enable_print_handler:
            handlers.append(PrintHandler())
        # yield the agent payload
        return dict(
            agent=agent,
            messages=messages,
            tools=as_tools(tools),
            handlers=handlers,
        )

    @prefect_task(task_run_name="Run LLM")
    async def run_once_async(self) -> list[MessageType]:
        async with self.tui():
            payload = self._setup_run()
            if payload is None:
                return
            agent: Agent = payload.pop("agent")
            response_handler = ResponseHandler()
            payload["handlers"].append(response_handler)

            with ctx(agent=agent, flow=self.flow, controller=self):
                response_gen = await completion_async(
                    messages=payload["messages"],
                    model=agent.get_model(),
                    tools=payload["tools"],
                    handlers=payload["handlers"],
                    max_iterations=1,
                    stream=True,
                    agent=agent,
                )
                async for _ in response_gen:
                    pass

            # save history
            self.flow.add_messages(
                messages=response_handler.response_messages,
            )
            self._iteration += 1

            create_messages_markdown_artifact(
                messages=response_handler.response_messages,
                thread_id=self.flow.thread_id,
            )

        return response_handler.response_messages

    @prefect_task(task_run_name="Run LLM")
    def run_once(self) -> list[MessageType]:
        payload = self._setup_run()
        if payload is None:
            return
        agent: Agent = payload.pop("agent")
        response_handler = ResponseHandler()
        payload["handlers"].append(response_handler)

        with ctx(
            agent=agent,
            flow=self.flow,
            controller=self,
        ):
            response_gen = completion(
                messages=payload["messages"],
                model=agent.get_model(),
                tools=payload["tools"],
                handlers=payload["handlers"],
                max_iterations=1,
                stream=True,
                agent=agent,
            )
            for _ in response_gen:
                pass

        # save history
        self.flow.add_messages(
            messages=response_handler.response_messages,
        )
        self._iteration += 1

        create_messages_markdown_artifact(
            messages=response_handler.response_messages,
            thread_id=self.flow.thread_id,
        )

        return response_handler.response_messages

    @prefect_task(task_run_name="Run LLM Controller")
    async def run_async(self) -> list[MessageType]:
        """
        Run the controller until all tasks are complete.
        """
        if all(t.is_complete() for t in self.tasks):
            return

        messages = []
        async with self.tui():
            # enter a flow context
            with self.flow:
                while (
                    any(t.is_incomplete() for t in self.tasks) and not self._should_stop
                ):
                    new_messages = await self.run_once_async()
                    if new_messages:
                        messages.extend(new_messages)
            self._should_stop = False
            return messages

    @prefect_task(task_run_name="Run LLM Controller")
    def run(self) -> list[MessageType]:
        """
        Run the controller until all tasks are complete.
        """
        if all(t.is_complete() for t in self.tasks):
            return

        messages = []
        # enter a flow context
        with self.flow:
            while any(t.is_incomplete() for t in self.tasks) and not self._should_stop:
                new_messages = self.run_once()
                if new_messages:
                    messages.extend(new_messages)
        self._should_stop = False
        return messages



import datetime
import uuid
from contextlib import contextmanager, nullcontext
from typing import Any, Callable, Optional, Union

from pydantic import Field

from controlflow.agents import Agent
from controlflow.flows.history import History, get_default_history
from controlflow.llm.messages import MessageType
from controlflow.tasks.task import Task
from controlflow.utilities.context import ctx
from controlflow.utilities.logging import get_logger
from controlflow.utilities.prefect import prefect_flow_context
from controlflow.utilities.types import ControlFlowModel

logger = get_logger(__name__)


class Flow(ControlFlowModel):
    name: Optional[str] = None
    description: Optional[str] = None
    thread_id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    history: History = Field(default_factory=get_default_history)
    tools: list[Callable] = Field(
        default_factory=list,
        description="Tools that will be available to every agent in the flow",
    )
    agents: list[Agent] = Field(
        description="The default agents for the flow. These agents will be used "
        "for any task that does not specify agents.",
        default_factory=list,
    )
    context: dict[str, Any] = {}
    tasks: dict[str, Task] = {}
    _cm_stack: list[contextmanager] = []

    def __init__(self, *, copy_parent: bool = True, **kwargs):
        """
        By default, the flow will copy the history from the parent flow if one
        exists, including all completed tasks. Because each flow is a new
        thread, new messages will not be shared between the parent and child
        flow.
        """
        super().__init__(**kwargs)
        parent = get_flow()
        if parent and copy_parent:
            self.add_messages(parent.get_messages())
            for task in parent.tasks.values():
                if task.is_complete():
                    self.add_task(task)

    def __enter__(self):
        # use stack so we can enter the context multiple times
        cm = self.create_context()
        self._cm_stack.append(cm)
        return cm.__enter__()

    def __exit__(self, *exc_info):
        # exit the context manager
        return self._cm_stack.pop().__exit__(*exc_info)

    def get_messages(
        self,
        limit: int = None,
        before: datetime.datetime = None,
        after: datetime.datetime = None,
    ) -> list[MessageType]:
        return self.history.load_messages(
            thread_id=self.thread_id, limit=limit, before=before, after=after
        )

    def add_messages(self, messages: list[MessageType]):
        self.history.save_messages(thread_id=self.thread_id, messages=messages)

    def add_task(self, task: Task):
        if self.tasks.get(task.id, task) is not task:
            raise ValueError(
                f"A different task with id '{task.id}' already exists in flow."
            )
        self.tasks[task.id] = task

    @contextmanager
    def create_context(self, create_prefect_flow_context: bool = True):
        ctx_args = dict(flow=self)
        if create_prefect_flow_context and ctx.get("prefect_flow") is not self:
            prefect_ctx = prefect_flow_context(name=self.name)
            ctx_args["prefect_flow"] = self
        else:
            prefect_ctx = nullcontext()

        with ctx(**ctx_args), prefect_ctx:
            yield self

    async def run_once_async(self):
        """
        Runs one step of the flow asynchronously.
        """
        if self.tasks:
            from controlflow.controllers import Controller

            controller = Controller(
                flow=self,
                tasks=list(self.tasks.values()),
            )
            await controller.run_once_async()

    def run_once(self):
        """
        Runs one step of the flow.
        """
        if self.tasks:
            from controlflow.controllers import Controller

            controller = Controller(
                flow=self,
                tasks=list(self.tasks.values()),
            )
            controller.run_once()

    async def run_async(self):
        """
        Runs the flow asynchronously.
        """
        if self.tasks:
            from controlflow.controllers import Controller

            controller = Controller(
                flow=self,
                tasks=list(self.tasks.values()),
            )
            await controller.run_async()

    def run(self):
        """
        Runs the flow.
        """
        if self.tasks:
            from controlflow.controllers import Controller

            controller = Controller(
                flow=self,
                tasks=list(self.tasks.values()),
            )
            controller.run()


def get_flow() -> Optional[Flow]:
    """
    Loads the flow from the context. If no flow is found, returns None.
    """
    flow: Union[Flow, None] = ctx.get("flow")
    return flow


def get_flow_messages(limit: int = None) -> list[MessageType]:
    """
    Loads messages from the flow's thread.

    Will error if no flow is found in the context.
    """
    if limit is None:
        limit = 50
    flow = get_flow()
    if flow:
        return get_default_history().load_messages(flow.thread_id, limit=limit)
    else:
        return []



import importlib

import controlflow
import pytest
from controlflow.settings import temporary_settings
from prefect.logging import get_logger


def test_temporary_settings():
    assert controlflow.settings.tools_raise_on_error is False
    with temporary_settings(tools_raise_on_error=True):
        assert controlflow.settings.tools_raise_on_error is True
    assert controlflow.settings.tools_raise_on_error is False


def test_prefect_settings_apply_at_runtime(caplog):
    prefect_logger = get_logger()
    assert controlflow.settings.prefect_log_level == "WARNING"
    prefect_logger.warning("test-log-1")
    controlflow.settings.prefect_log_level = "ERROR"
    prefect_logger.warning("test-log-2")
    controlflow.settings.prefect_log_level = "DEBUG"
    prefect_logger.warning("test-log-3")

    assert "test-log-1" in caplog.text
    assert "test-log-2" not in caplog.text
    assert "test-log-3" in caplog.text


def test_import_without_default_api_key_warns_but_does_not_fail(monkeypatch, caplog):
    # remove the OPENAI_API_KEY environment variable
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)

    # Clear any previous logs
    caplog.clear()

    # Import the library
    with caplog.at_level("WARNING"):
        # Reload the library to apply changes
        importlib.reload(controlflow)

    # Check if the warning was logged
    assert any(
        record.levelname == "WARNING"
        and "The default LLM model could not be created" in record.message
        for record in caplog.records
    ), "The expected warning was not logged"


def test_import_without_default_api_key_errors_when_loading_model(monkeypatch):
    # remove the OPENAI_API_KEY environment variable
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)

    # Reload the library to apply changes
    importlib.reload(controlflow)

    with pytest.raises(ValueError, match="Did not find openai_api_key"):
        controlflow.get_default_model()

    with pytest.raises(
        ValueError, match="No model provided and no default model could be loaded"
    ):
        controlflow.Agent().get_model()


def test_import_without_api_key_for_non_default_model_warns_but_does_not_fail(
    monkeypatch, caplog
):
    # remove the OPENAI_API_KEY environment variable
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)
    monkeypatch.setenv("CONTROLFLOW_LLM_MODEL", "anthropic/not-a-model")

    # Clear any previous logs
    caplog.clear()

    # Import the library
    with caplog.at_level("WARNING"):
        # Reload the library to apply changes
        importlib.reload(controlflow)

    # Check if the warning was logged
    assert any(
        record.levelname == "WARNING"
        and "The default LLM model could not be created" in record.message
        for record in caplog.records
    ), "The expected warning was not logged"



import math
from typing import TYPE_CHECKING, AsyncGenerator, Callable, Generator, Optional, Union

import langchain_core.language_models as lc_models
import langchain_core.messages
import tiktoken
from langchain_core.messages.utils import trim_messages

import controlflow
import controlflow.llm.models
from controlflow.llm.handlers import (
    CompletionEvent,
    CompletionHandler,
    ResponseHandler,
)
from controlflow.llm.messages import AIMessage, AIMessageChunk, BaseMessage, MessageType
from controlflow.llm.tools import ToolCall, as_tools, handle_tool_call

if TYPE_CHECKING:
    from controlflow.agents.agent import Agent


def token_counter(message: langchain_core.messages.BaseMessage) -> int:
    # always use gpt-3.5 token counter with the entire message object; we only need to be approximate here
    return len(tiktoken.encoding_for_model("gpt-3.5-turbo").encode(message.json()))


def handle_tool_calls(
    message: AIMessage,
    tools: list[Callable],
    response_messages: list[MessageType],
    agent: Optional["Agent"] = None,
):
    """
    Emit events for the given message when it has tool calls.
    """
    for tool_call in message.tool_calls:
        yield CompletionEvent(
            type="tool_result_created",
            payload=dict(message=message, tool_call=tool_call),
        )
        error = handle_multiple_talk_to_user_calls(tool_call, message)
        tool_result_message = handle_tool_call(
            tool_call, tools, error=error, agent=agent
        )
        response_messages.append(tool_result_message)
        yield CompletionEvent(
            type="tool_result_done", payload=dict(message=tool_result_message)
        )


def handle_delta_events(
    delta: langchain_core.messages.AIMessageChunk,
    snapshot: langchain_core.messages.AIMessageChunk,
    deltas: list[langchain_core.messages.AIMessageChunk],
    agent: "Agent",
):
    """
    Emit events for the given delta message.

    Note this function receives langchain messages
    """
    delta = AIMessageChunk.from_langchain_message(delta, agent=agent)
    snapshot = AIMessageChunk.from_langchain_message(snapshot, agent=agent)

    if delta.content:
        if not deltas[-1].content:
            yield CompletionEvent(type="message_created", payload=dict(delta=delta))
        if delta.content != deltas[-1].content:
            yield CompletionEvent(
                type="message_delta",
                payload=dict(delta=delta, snapshot=snapshot),
            )

    if delta.tool_calls:
        if not deltas[-1].tool_calls:
            yield CompletionEvent(type="tool_call_created", payload=dict(delta=delta))
        yield CompletionEvent(
            type="tool_call_delta",
            payload=dict(delta=delta, snapshot=snapshot),
        )


def handle_done_events(message: AIMessage):
    """
    Emit events for the given message when it has been processed.
    """
    if message.content:
        yield CompletionEvent(type="message_done", payload=dict(message=message))
    if message.tool_calls:
        yield CompletionEvent(type="tool_call_done", payload=dict(message=message))


def handle_multiple_talk_to_user_calls(tool_call: ToolCall, message: AIMessage):
    if (
        tool_call["name"] == "talk_to_user"
        and len([t for t in message.tool_calls if t["name"] == "talk_to_user"]) > 1
    ):
        error = 'Tool call "talk_to_user" can only be used once per turn.'
    else:
        error = None
    return error


def prepare_messages(messages: list[MessageType]) -> list[MessageType]:
    """
    Make any necessary modifications to the messages before they are passed to the model.
    """
    return messages


def _completion_generator(
    messages: list[MessageType],
    model: lc_models.BaseChatModel,
    tools: Optional[list[Callable]],
    max_iterations: int,
    stream: bool,
    agent: Optional["Agent"] = None,
    **kwargs,
) -> Generator[CompletionEvent, None, None]:
    response_messages = []
    response_message = None

    if tools:
        model = model.bind_tools([t.to_lc_tool() for t in as_tools(tools)])

    counter = 0
    try:
        yield CompletionEvent(type="start", payload={})

        # continue as long as the last response message contains tool calls (or
        # there is no response message yet)
        while not response_message or response_message.tool_calls:
            input_messages = [
                m.to_langchain_message()
                for m in messages + response_messages
                if isinstance(m, BaseMessage)
            ]
            input_messages = trim_messages(
                messages=input_messages,
                max_tokens=controlflow.settings.max_input_tokens,
                include_system=True,
                token_counter=token_counter,
            )

            if not stream:
                response_message = model.invoke(input=input_messages, **kwargs)
                response_message = AIMessage.from_langchain_message(
                    response_message, agent=agent
                )

            else:
                # all streaming responses are langchain Pydantic v1 models
                # which we don't convert to AIMessage/AIMessageChunks for sanity.
                # they are converted in handle_delta_events and when the stream is finished.

                # initialize the list of deltas with an empty delta
                # to facilitate comparison with the previous delta
                deltas = [langchain_core.messages.AIMessageChunk(content="")]

                for i, delta in enumerate(model.stream(input=input_messages, **kwargs)):
                    if i == 0:
                        snapshot = delta
                    else:
                        snapshot = snapshot + delta

                    yield from handle_delta_events(
                        delta=delta, snapshot=snapshot, deltas=deltas, agent=agent
                    )

                    deltas.append(delta)

                # the last snapshot message is the response message
                response_message = AIMessage.from_langchain_message(
                    snapshot, agent=agent
                )

            # handle done events for the response message
            yield from handle_done_events(response_message)

            # append the response message to the list of response messages
            response_messages.append(response_message)

            # handle tool calls
            yield from handle_tool_calls(
                message=response_message,
                tools=tools,
                response_messages=response_messages,
                agent=agent,
            )

            counter += 1
            if counter >= (max_iterations or math.inf):
                break

    except (BaseException, Exception) as exc:
        yield CompletionEvent(type="exception", payload=dict(exc=exc))
        raise
    finally:
        yield CompletionEvent(type="end", payload={})


async def _completion_async_generator(
    messages: list[MessageType],
    model: lc_models.BaseChatModel,
    tools: Optional[list[Callable]],
    max_iterations: int,
    stream: bool,
    agent: Optional["Agent"] = None,
    **kwargs,
) -> AsyncGenerator[CompletionEvent, None]:
    response_messages = []
    response_message = None

    if tools:
        model = model.bind_tools([t.to_lc_tool() for t in as_tools(tools)])

    counter = 0
    try:
        yield CompletionEvent(type="start", payload={})

        # continue as long as the last response message contains tool calls (or
        # there is no response message yet)
        while not response_message or response_message.tool_calls:
            input_messages = [
                m.to_langchain_message() if isinstance(m, BaseMessage) else m
                for m in messages + response_messages
            ]

            input_messages = trim_messages(
                messages=input_messages,
                max_tokens=controlflow.settings.max_input_tokens,
                include_system=True,
                token_counter=token_counter,
            )
            input_messages = [
                m.to_langchain_message()
                for m in input_messages
                if isinstance(m, BaseMessage)
            ]

            if not stream:
                response_message = await model.ainvoke(input=input_messages, **kwargs)
                response_message = AIMessage.from_langchain_message(
                    response_message, agent=agent
                )

            else:
                # all streaming responses are langchain Pydantic v1 models
                # which we don't convert to AIMessage/AIMessageChunks for sanity.
                # they are converted in handle_delta_events and when the stream is finished.

                # initialize the list of deltas with an empty delta
                # to facilitate comparison with the previous delta
                deltas = [langchain_core.messages.AIMessageChunk(content="")]

                async for i, delta in enumerate(
                    model.astream(input=input_messages, **kwargs)
                ):
                    if i == 0:
                        snapshot = delta
                    else:
                        snapshot = snapshot + delta

                    for event in handle_delta_events(
                        delta=delta, snapshot=snapshot, deltas=deltas, agent=agent
                    ):
                        yield event

                    deltas.append(delta)

                # the last snapshot message is the response message
                response_message = AIMessage.from_langchain_message(
                    snapshot, agent=agent
                )

            # handle done events for the response message
            for event in handle_done_events(response_message):
                yield event

            # append the response message to the list of response messages
            response_messages.append(response_message)

            # handle tool calls
            for event in handle_tool_calls(
                message=response_message,
                tools=tools,
                response_messages=response_messages,
                agent=agent,
            ):
                yield event

            counter += 1
            if counter >= (max_iterations or math.inf):
                break

    except (BaseException, Exception) as exc:
        yield CompletionEvent(type="exception", payload=dict(exc=exc))
        raise
    finally:
        yield CompletionEvent(type="end", payload={})


def _handle_events(
    generator: Generator[CompletionEvent, None, None], handlers: list[CompletionHandler]
) -> Generator[CompletionEvent, None, None]:
    for event in generator:
        for handler in handlers:
            try:
                handler.on_event(event)
            except Exception as exc:
                generator.throw(exc)
        yield event


async def _handle_events_async(
    generator: AsyncGenerator, handlers: list[CompletionHandler]
) -> AsyncGenerator[CompletionEvent, None]:
    async for event in generator:
        for handler in handlers:
            try:
                handler.on_event(event)
            except Exception as exc:
                await generator.athrow(exc)
        yield event


def completion(
    messages: list[MessageType],
    model: lc_models.BaseChatModel = None,
    tools: list[Callable] = None,
    max_iterations: int = None,
    handlers: list[CompletionHandler] = None,
    stream: bool = False,
    agent: Optional["Agent"] = None,
    **kwargs,
) -> Union[list[MessageType], Generator[MessageType, None, None]]:
    if model is None:
        model = controlflow.llm.models.get_default_model()

    response_handler = ResponseHandler()
    handlers = handlers or []
    handlers.append(response_handler)

    completion_generator = _completion_generator(
        messages=messages,
        model=model,
        tools=tools,
        max_iterations=max_iterations,
        stream=stream,
        agent=agent,
        **kwargs,
    )

    handlers_generator = _handle_events(completion_generator, handlers)

    if stream:
        return handlers_generator
    else:
        for _ in handlers_generator:
            pass
        return response_handler.response_messages


async def completion_async(
    messages: list[MessageType],
    model: lc_models.BaseChatModel = None,
    tools: list[Callable] = None,
    max_iterations: int = None,
    handlers: list[CompletionHandler] = None,
    stream: bool = False,
    agent: Optional["Agent"] = None,
    **kwargs,
) -> Union[list[MessageType], Generator[MessageType, None, None]]:
    if model is None:
        model = controlflow.llm.models.get_default_model()

    response_handler = ResponseHandler()
    handlers = handlers or []
    handlers.append(response_handler)

    completion_generator = _completion_async_generator(
        messages=messages,
        model=model,
        tools=tools,
        max_iterations=max_iterations,
        stream=stream,
        agent=agent,
        **kwargs,
    )

    handlers_generator = _handle_events_async(completion_generator, handlers)

    if stream:
        return handlers_generator
    else:
        async for _ in handlers_generator:
            pass
        return response_handler.response_messages



from controlflow.agents import Agent
from controlflow.flows import Flow, get_flow_messages
from controlflow.instructions import get_instructions
from controlflow.llm.classify import classify
from controlflow.tasks.task import Task


def round_robin(
    agents: list[Agent],
    task: Task,
    flow: Flow,
    **kwargs,
) -> Agent:
    return agents[task._iteration % len(agents)]


def moderator(
    agents: list[Agent],
    tasks: list[Task],
    context: dict = None,
    iteration: int = 0,
    model: str = None,
) -> Agent:
    history = get_flow_messages()
    instructions = get_instructions()
    context = context or {}
    context.update(tasks=tasks, history=history, instructions=instructions)

    agent = classify(
        context,
        labels=agents,
        instructions="""
            Given the context, choose the AI agent best suited to take the
            next turn at completing the tasks in the task graph. Take into account
            any descriptions, tasks, history, instructions, and tools. Focus on
            agents assigned to upstream dependencies or subtasks that need to be
            completed before their downstream/parents can be completed. An agent
            can only work on a task that it is assigned to.
            """,
        model=model,
    )
    return agent



import logging
import random
import uuid
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Optional

from langchain_core.language_models import BaseChatModel
from pydantic import Field, field_serializer

import controlflow
from controlflow.instructions import get_instructions
from controlflow.llm.models import get_default_model
from controlflow.llm.rules import LLMRules
from controlflow.tools.talk_to_user import talk_to_user
from controlflow.utilities.context import ctx
from controlflow.utilities.types import ControlFlowModel

from .memory import Memory
from .names import NAMES

if TYPE_CHECKING:
    from controlflow.tasks.task import Task
logger = logging.getLogger(__name__)


def get_default_agent() -> "Agent":
    return controlflow.default_agent


class Agent(ControlFlowModel):
    model_config = dict(arbitrary_types_allowed=True)
    id: str = Field(default_factory=lambda: str(uuid.uuid4().hex[:5]))
    name: str = Field(
        description="The name of the agent.",
        default_factory=lambda: random.choice(NAMES),
    )
    description: Optional[str] = Field(
        None, description="A description of the agent, visible to other agents."
    )
    instructions: Optional[str] = Field(
        "You are a diligent AI assistant. You complete your tasks efficiently and without error.",
        description="Instructions for the agent, private to this agent.",
    )
    tools: list[Callable] = Field(
        [], description="List of tools availble to the agent."
    )
    user_access: bool = Field(
        False,
        description="If True, the agent is given tools for interacting with a human user.",
    )
    memory: Optional[Memory] = Field(
        default=None,
        # default_factory=ThreadMemory,
        description="The memory object used by the agent. If not specified, an in-memory memory object will be used. Pass None to disable memory.",
    )

    # note: `model` should be typed as Optional[BaseChatModel] but V2 models can't have
    # V1 attributes without erroring, so we have to use Any.
    model: Optional[Any] = Field(
        None,
        description="The LangChain BaseChatModel used by the agent. If not provided, the default model will be used.",
        exclude=True,
    )

    _cm_stack: list[contextmanager] = []

    @field_serializer("tools")
    def _serialize_tools(self, tools: list[Callable]):
        tools = controlflow.tools.as_tools(tools)
        # tools are Pydantic 1 objects
        return [t.dict(include={"name", "description"}) for t in tools]

    def __init__(self, name=None, **kwargs):
        if name is not None:
            kwargs["name"] = name

        if additional_instructions := get_instructions():
            kwargs["instructions"] = (
                kwargs.get("instructions")
                or "" + "\n" + "\n".join(additional_instructions)
            ).strip()

        super().__init__(**kwargs)

    def serialize_for_prompt(self) -> dict:
        dct = self.model_dump(
            include={"name", "id", "description", "tools", "user_access"}
        )
        # seeing user access = False can confuse agents on tasks with user access
        if not dct["user_access"]:
            dct.pop("user_access")
        return dct

    def get_model(self) -> BaseChatModel:
        """
        Retrieve the LLM model for this agent
        """
        try:
            return self.model or get_default_model()
        except Exception as exc:
            raise ValueError(
                f"Agent {self.name}: No model provided and no default model could be loaded: {exc}"
            ) from exc

    def get_llm_rules(self) -> LLMRules:
        """
        Retrieve the LLM rules for this agent's model
        """
        return controlflow.llm.rules.rules_for_model(self.get_model())

    def get_tools(self) -> list[Callable]:
        tools = self.tools.copy()
        if self.user_access:
            tools.append(talk_to_user)
        if self.memory is not None:
            tools.extend(self.memory.get_tools())
        return tools

    @contextmanager
    def create_context(self):
        with ctx(agent=self):
            yield self

    def __enter__(self):
        # use stack so we can enter the context multiple times
        self._cm_stack.append(self.create_context())
        return self._cm_stack[-1].__enter__()

    def __exit__(self, *exc_info):
        return self._cm_stack.pop().__exit__(*exc_info)

    def run_once(self, task: "Task"):
        return task.run_once(agents=[self])

    async def run_once_async(self, task: "Task"):
        return await task.run_once_async(agents=[self])

    def run(self, task: "Task"):
        return task.run(agents=[self])

    async def run_async(self, task: "Task"):
        return await task.run_async(agents=[self])


DEFAULT_AGENT = Agent(name="Marvin")



import controlflow
from controlflow.llm.messages import AgentReference, AIMessage


class TestAIMessage:
    def test_agent(self):
        agent = controlflow.Agent(name="Test Agent!")
        message = AIMessage(content="", agent=agent)
        assert isinstance(message.agent, AgentReference)
        assert message.agent.name == "Test Agent!"

    def test_name_loaded_from_agent(self):
        agent = controlflow.Agent(name="Test Agent!")
        message = AIMessage(content="", agent=agent)
        assert message.name == "Test-Agent"



from langchain_anthropic import ChatAnthropic
from langchain_openai import AzureChatOpenAI, ChatOpenAI

from controlflow.llm.models import BaseChatModel


class LLMRules:
    """
    LLM rules let us tailor DAG compilation, message generation, tool use, and
    other behavior to the requirements of different LLM provider APIs.

    Rules can be added here (to the base class) and overridden in subclasses, if
    necessary.
    """

    # system messages can only be provided as the very first message in a thread
    system_message_must_be_first: bool = False

    # other than a system message, the first message must be from the user
    user_message_must_be_first_after_system: bool = False

    # the last message in a thread can't be from an AI if tool use is allowed
    allow_last_message_has_ai_role_with_tools: bool = True

    # consecutive AI messages must be separated by a user message
    allow_consecutive_ai_messages: bool = True

    # add system messages to identify speakers in multi-agent conversations
    # (some APIs can use the `name` field for this purpose, but others can't)
    add_system_messages_for_multi_agent: bool = False


class OpenAIRules(LLMRules):
    pass


class AnthropicRules(LLMRules):
    system_message_must_be_first: bool = True
    user_message_must_be_first_after_system: bool = True
    allow_last_message_has_ai_role_with_tools: bool = False
    allow_consecutive_ai_messages: bool = False


def rules_for_model(model: BaseChatModel) -> LLMRules:
    if isinstance(model, (ChatOpenAI, AzureChatOpenAI)):
        return OpenAIRules()
    elif isinstance(model, ChatAnthropic):
        return AnthropicRules()
    else:
        return LLMRules()



import pytest
from controlflow import Agent, Task, flow

pytest.skip("Skipping the entire file", allow_module_level=True)

# define assistants
user_agent = Agent(name="user-agent", user_access=True)
non_user_agent = Agent(name="non-user-agent", user_access=False)


def test_no_user_access_fails():
    @flow
    def user_access_flow():
        task = Task(
            "This task requires human user access. Inform the user that today is a good day.",
            agents=[non_user_agent],
        )
        task.run()

    with pytest.raises(ValueError):
        user_access_flow()


def test_user_access_agent_succeeds():
    @flow
    def user_access_flow():
        task = Task(
            "This task requires human user access. Inform the user that today is a good day.",
            agents=[user_agent],
        )
        task.run()

    assert user_access_flow()


def test_user_access_task_succeeds():
    @flow
    def user_access_flow():
        task = Task(
            "This task requires human user access. Inform the user that today is a good day.",
            agents=[non_user_agent],
            user_access=True,
        )
        task.run()

    assert user_access_flow()


def test_user_access_agent_and_task_succeeds():
    @flow
    def user_access_flow():
        task = Task(
            "This task requires human user access. Inform the user that today is a good day.",
            agents=[user_agent],
            user_access=True,
        )
        task.run()

    assert user_access_flow()



from typing import Optional, TypeVar, Union

from pydantic import Field

from controlflow.agents import Agent
from controlflow.flows import Flow
from controlflow.tasks.task import Task
from controlflow.tools import Tool, as_tools
from controlflow.utilities.types import ControlFlowModel

ToolLiteral = TypeVar("ToolLiteral", bound=str)


class PlanTask(ControlFlowModel):
    id: int
    objective: str
    instructions: Optional[str] = Field(
        None,
        description="Any additional instructions for completing the task objective.",
    )
    depends_on: list[int] = Field(
        [], description="Tasks that must be completed before this task can be started."
    )
    parent: Optional[int] = Field(None, description="The parent of this task (if any).")
    agents: list[int] = Field(
        [],
        description="The agents assigned to the task. If empty, the default agent is used.",
    )
    tools: list[int] = Field(
        [],
        description="The tools provided to complete the task. If empty, no tools are provided.",
    )


def create_plan(
    objective: str,
    instructions: str = None,
    planning_agent: Agent = None,
    agents: list[Agent] = None,
    tools: list[Union[callable, Tool]] = None,
    context: dict = None,
) -> list[Task]:
    """
    Given an objective and instructions for achieving it, generate a plan for
    completing the objective. Each step of the plan will be turned into a task
    objective.
    """
    tools = as_tools(tools or [])

    agent_dict = dict(enumerate(agents or []))
    tool_dict = dict(
        enumerate([t.dict(include={"name", "description"}) for t in tools])
    )

    task = Task(
        objective="""
            Create a plan consisting of multiple tasks to complete the provided objective.
            """,
        instructions="""
            Use your mark_successful tool to create the plan. Do not post a
            message or talk out loud.

            Each task should be a discrete, actionable step that contributes to
            the overall objective. Do not waste time on uneccessary or redundant
            steps. Make sure to use your tools. 
            
            When creating tasks, imagine that you had to complete the plan
            yourself. What steps would you take? What tools would you use? What
            information would you need? Remember that each task has a token cost
            (both in its evaluation and needing to mark it complete), so try to
            organize objectives by outcomes and dependencies, not by the actions
            you'd need to take.
            
            - Use `depends_on` to indicate which tasks must be completed before
              others can start. Tasks can only depend on tasks that come before
              them in your plan. 
            - Use `parent` to indicate tasks that are subtasks of others.
            - Don't create needless tasks like "document the findings"
            
        """,
        context=dict(
            plan_objective=objective,
            plan_instructions=instructions,
            plan_agents=agent_dict,
            plan_tools=tool_dict,
        )
        | context
        or {},
        agents=[planning_agent] if planning_agent else None,
        result_type=list[PlanTask],
        infer_parent=False,
    )

    # create a new flow to avoid polluting the main flow's history
    with Flow():
        task.run()

    plan: list[PlanTask] = task.result

    task_ids = {}

    for t in plan:
        try:
            task_agents = [agent_dict[i] for i in t.agents] if t.agents else None
        except KeyError:
            task_agents = None

        try:
            task_tools = [tool_dict[i] for i in t.tools]
        except KeyError:
            task_tools = []

        task_ids[t.id] = Task(
            objective=t.objective,
            instructions=t.instructions,
            depends_on=[task_ids[i] for i in t.depends_on],
            parent=task_ids[t.parent] if t.parent else None,
            agents=task_agents,
            tools=task_tools,
            context=context,
        )

    return list(task_ids.values())



from functools import partial

import pytest
from controlflow.agents import Agent, get_default_agent
from controlflow.controllers.graph import EdgeType
from controlflow.flows import Flow
from controlflow.instructions import instructions
from controlflow.tasks.task import Task, TaskStatus
from controlflow.utilities.context import ctx

SimpleTask = partial(Task, objective="test", result_type=None)


def test_context_open_and_close():
    assert ctx.get("tasks") == []
    with SimpleTask() as ta:
        assert ctx.get("tasks") == [ta]
        with SimpleTask() as tb:
            assert ctx.get("tasks") == [ta, tb]
        assert ctx.get("tasks") == [ta]
    assert ctx.get("tasks") == []


def test_task_requires_objective():
    with pytest.raises(ValueError):
        Task()


def test_task_initialization():
    task = Task(objective="Test objective")
    assert task.objective == "Test objective"
    assert task.status == TaskStatus.INCOMPLETE
    assert task.result is None
    assert task.error is None


def test_task_loads_instructions_at_creation():
    with instructions("test instruction"):
        task = SimpleTask()

    assert "test instruction" in task.instructions


def test_task_dependencies():
    task1 = SimpleTask()
    task2 = SimpleTask(depends_on=[task1])
    assert task1 in task2.depends_on
    assert task2 in task1._downstreams


def test_task_subtasks():
    task1 = SimpleTask()
    task2 = SimpleTask(parent=task1)
    assert task2 in task1._subtasks
    assert task2.parent is task1


def test_task_parent_context():
    with SimpleTask() as task1:
        with SimpleTask() as task2:
            task3 = SimpleTask()

    assert task3.parent is task2
    assert task2.parent is task1
    assert task1.parent is None

    assert task1._subtasks == {task2}
    assert task2._subtasks == {task3}
    assert task3._subtasks == set()


def test_task_agent_assignment():
    agent = Agent(name="Test Agent")
    task = SimpleTask(agents=[agent])
    assert agent in task.agents


def test_task_bad_agent_assignment():
    with pytest.raises(ValueError):
        SimpleTask(agents=[])


def test_task_loads_agent_from_parent():
    agent = Agent(name="Test Agent")
    with SimpleTask(agents=[agent]):
        child = SimpleTask()

    assert child.agents is None
    assert child.get_agents() == [agent]


def test_task_loads_agent_from_flow():
    def_agent = get_default_agent()
    agent = Agent(name="Test Agent")
    with Flow(agents=[agent]):
        task = SimpleTask()

        assert task.agents is None
        assert task.get_agents() == [agent]

    # outside the flow context, pick up the default agent
    assert task.get_agents() == [def_agent]


def test_task_loads_agent_from_default_if_none_otherwise():
    agent = get_default_agent()
    task = SimpleTask()

    assert task.agents is None
    assert task.get_agents() == [agent]


def test_task_loads_agent_from_parent_before_flow():
    agent1 = Agent(name="Test Agent 1")
    agent2 = Agent(name="Test Agent 2")
    with Flow(agents=[agent1]):
        with SimpleTask(agents=[agent2]):
            child = SimpleTask()

    assert child.agents is None
    assert child.get_agents() == [agent2]


def test_task_tracking():
    with Flow() as flow:
        task = SimpleTask()
        assert task in flow.tasks.values()


def test_task_tracking_on_call():
    task = SimpleTask()
    with Flow() as flow:
        task.run_once()
    assert task in flow.tasks.values()


class TestTaskStatus:
    def test_task_status_transitions(self):
        task = SimpleTask()
        assert task.is_incomplete()
        assert not task.is_complete()
        assert not task.is_successful()
        assert not task.is_failed()
        assert not task.is_skipped()

        task.mark_successful()
        assert not task.is_incomplete()
        assert task.is_complete()
        assert task.is_successful()
        assert not task.is_failed()
        assert not task.is_skipped()

        task = SimpleTask()
        task.mark_failed()
        assert not task.is_incomplete()
        assert task.is_complete()
        assert not task.is_successful()
        assert task.is_failed()
        assert not task.is_skipped()

        task = SimpleTask()
        task.mark_skipped()
        assert not task.is_incomplete()
        assert task.is_complete()
        assert not task.is_successful()
        assert not task.is_failed()
        assert task.is_skipped()

    def test_validate_upstream_dependencies_on_success(self):
        task1 = SimpleTask()
        task2 = SimpleTask(depends_on=[task1])
        with pytest.raises(ValueError, match="cannot be marked successful"):
            task2.mark_successful()
        task1.mark_successful()
        task2.mark_successful()

    def test_validate_subtask_dependencies_on_success(self):
        task1 = SimpleTask()
        task2 = SimpleTask(parent=task1)
        with pytest.raises(ValueError, match="cannot be marked successful"):
            task1.mark_successful()
        task2.mark_successful()
        task1.mark_successful()

    def test_task_ready(self):
        task1 = SimpleTask()
        assert task1.is_ready()

    def test_task_not_ready_if_successful(self):
        task1 = SimpleTask()
        task1.mark_successful()
        assert not task1.is_ready()

    def test_task_not_ready_if_failed(self):
        task1 = SimpleTask()
        task1.mark_failed()
        assert not task1.is_ready()

    def test_task_not_ready_if_dependencies_are_ready(self):
        task1 = SimpleTask()
        task2 = SimpleTask(depends_on=[task1])
        assert task1.is_ready()
        assert not task2.is_ready()

    def test_task_ready_if_dependencies_are_ready(self):
        task1 = SimpleTask()
        task2 = SimpleTask(depends_on=[task1])
        task1.mark_successful()
        assert not task1.is_ready()
        assert task2.is_ready()

    def test_task_hash(self):
        task1 = SimpleTask()
        task2 = SimpleTask()
        assert hash(task1) != hash(task2)

    def test_ready_task_adds_tools(self):
        task = SimpleTask()
        assert task.is_ready()

        tools = task.get_tools()
        assert any(tool.name == f"mark_task_{task.id}_failed" for tool in tools)
        assert any(tool.name == f"mark_task_{task.id}_successful" for tool in tools)

    def test_completed_task_does_not_add_tools(self):
        task = SimpleTask()
        task.mark_successful()
        tools = task.get_tools()
        assert not any(tool.name == f"mark_task_{task.id}_failed" for tool in tools)
        assert not any(tool.name == f"mark_task_{task.id}_successful" for tool in tools)

    def test_task_with_incomplete_upstream_does_not_add_tools(self):
        upstream_task = SimpleTask()
        downstream_task = SimpleTask(depends_on=[upstream_task])
        tools = downstream_task.get_tools()
        assert not any(
            tool.name == f"mark_task_{downstream_task.id}_failed" for tool in tools
        )
        assert not any(
            tool.name == f"mark_task_{downstream_task.id}_successful" for tool in tools
        )

    def test_task_with_incomplete_subtask_does_not_add_tools(self):
        parent = SimpleTask()
        SimpleTask(parent=parent)
        tools = parent.get_tools()
        assert not any(tool.name == f"mark_task_{parent.id}_failed" for tool in tools)
        assert not any(
            tool.name == f"mark_task_{parent.id}_successful" for tool in tools
        )


class TestTaskToGraph:
    def test_single_task_graph(self):
        task = SimpleTask()
        graph = task.as_graph()
        assert len(graph.tasks) == 1
        assert task in graph.tasks
        assert len(graph.edges) == 0

    def test_task_with_subtasks_graph(self):
        task1 = SimpleTask()
        task2 = SimpleTask(parent=task1)
        graph = task1.as_graph()
        assert len(graph.tasks) == 2
        assert task1 in graph.tasks
        assert task2 in graph.tasks
        assert len(graph.edges) == 1
        assert any(
            edge.upstream == task2
            and edge.downstream == task1
            and edge.type == EdgeType.SUBTASK
            for edge in graph.edges
        )

    def test_task_with_dependencies_graph(self):
        task1 = SimpleTask()
        task2 = SimpleTask(depends_on=[task1])
        graph = task2.as_graph()
        assert len(graph.tasks) == 2
        assert task1 in graph.tasks
        assert task2 in graph.tasks
        assert len(graph.edges) == 1
        assert any(
            edge.upstream == task1
            and edge.downstream == task2
            and edge.type == EdgeType.DEPENDENCY
            for edge in graph.edges
        )

    def test_task_with_subtasks_and_dependencies_graph(self):
        task1 = SimpleTask()
        task2 = SimpleTask(depends_on=[task1])
        task3 = SimpleTask(objective="Task 3", parent=task2)
        graph = task2.as_graph()
        assert len(graph.tasks) == 3
        assert task1 in graph.tasks
        assert task2 in graph.tasks
        assert task3 in graph.tasks
        assert len(graph.edges) == 2
        assert any(
            edge.upstream == task1
            and edge.downstream == task2
            and edge.type == EdgeType.DEPENDENCY
            for edge in graph.edges
        )
        assert any(
            edge.upstream == task3
            and edge.downstream == task2
            and edge.type == EdgeType.SUBTASK
            for edge in graph.edges
        )



import abc
import datetime
import json
import math
from pathlib import Path

from pydantic import Field, field_validator

import controlflow
from controlflow.llm.messages import MessageType
from controlflow.utilities.types import ControlFlowModel

# This is a global variable that will be shared between all instances of InMemoryHistory
IN_MEMORY_HISTORY = {}


def get_default_history() -> "History":
    return controlflow.default_history


class History(ControlFlowModel, abc.ABC):
    @abc.abstractmethod
    def load_messages(
        self,
        thread_id: str,
        limit: int = None,
        before: datetime.datetime = None,
        after: datetime.datetime = None,
    ) -> list[MessageType]:
        raise NotImplementedError()

    @abc.abstractmethod
    def save_messages(self, thread_id: str, messages: list[MessageType]):
        raise NotImplementedError()


class InMemoryHistory(History):
    history: dict[str, list[MessageType]] = Field(
        default_factory=lambda: IN_MEMORY_HISTORY
    )

    def load_messages(
        self,
        thread_id: str,
        limit: int = None,
        before: datetime.datetime = None,
        after: datetime.datetime = None,
    ) -> list[MessageType]:
        messages = self.history.get(thread_id, [])
        filtered_messages = [
            msg
            for i, msg in enumerate(reversed(messages))
            if (before is None or msg.timestamp < before)
            and (after is None or msg.timestamp > after)
            and i < (limit or math.inf)
        ]
        return list(reversed(filtered_messages))

    def save_messages(self, thread_id: str, messages: list[MessageType]):
        self.history.setdefault(thread_id, []).extend(messages)


class FileHistory(History):
    base_path: Path = Field(
        default_factory=lambda: controlflow.settings.home_path / "history"
    )

    def path(self, thread_id: str) -> Path:
        return self.base_path / f"{thread_id}.json"

    @field_validator("base_path", mode="before")
    def _validate_path(cls, v):
        v = Path(v).expanduser()
        if not v.exists():
            v.mkdir(parents=True, exist_ok=True)
        return v

    def load_messages(
        self,
        thread_id: str,
        limit: int = None,
        before: datetime.datetime = None,
        after: datetime.datetime = None,
    ) -> list[MessageType]:
        if not self.path(thread_id).exists():
            return []

        with open(self.path(thread_id), "r") as f:
            all_messages = json.load(f)

        messages = []
        for msg in reversed(all_messages):
            message = MessageType.model_validate(msg)
            if before is None or message.timestamp < before:
                if after is None or message.timestamp > after:
                    messages.append(message)
                    if len(messages) >= limit or math.inf:
                        break

        return list(reversed(messages))

    def save_messages(self, thread_id: str, messages: list[MessageType]):
        if self.path(thread_id).exists():
            with open(self.path(thread_id), "r") as f:
                all_messages = json.load(f)
        else:
            all_messages = []
        all_messages.extend([msg.model_dump(mode="json") for msg in messages])
        with open(self.path(thread_id), "w") as f:
            json.dump(all_messages, f)



import datetime
import re
import uuid
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

import langchain_core.messages
from langchain_core.messages.tool import InvalidToolCall, ToolCall, ToolCallChunk
from pydantic import Field, field_validator

from controlflow.utilities.jinja import jinja_env
from controlflow.utilities.types import ControlFlowModel

if TYPE_CHECKING:
    from controlflow.agents.agent import Agent


class BaseMessage(ControlFlowModel):
    """
    ControlFlow uses Message objects that are similar to LangChain messages, but more purpose built.

    Note that LangChain messages are Pydantic V1 models, while ControlFlow messages are Pydantic V2 models.
    """

    id: Optional[str] = Field(default_factory=lambda: uuid.uuid4().hex)
    timestamp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
    )
    role: str
    content: Union[str, list[Union[str, dict]]]
    name: Optional[str] = None

    # private attr to hold the original langchain message
    _langchain_message: Optional[
        Union[
            langchain_core.messages.BaseMessage,
            langchain_core.messages.BaseMessageChunk,
        ]
    ] = None

    @field_validator("name")
    def _sanitize_name(cls, v):
        # sanitize name for API compatibility - OpenAI API only allows alphanumeric characters, dashes, and underscores
        if v is not None:
            v = re.sub(r"[^a-zA-Z0-9_-]", "-", v).strip("-")
        return v

    def __init__(
        self,
        content: Union[str, list[Union[str, dict]]],
        _langchain_message: Optional[
            Union[
                langchain_core.messages.BaseMessage,
                langchain_core.messages.BaseMessageChunk,
            ]
        ] = None,
        **kwargs: Any,
    ) -> None:
        """Pass in content as positional arg."""
        super().__init__(content=content, **kwargs)
        self._langchain_message = _langchain_message

    @property
    def str_content(self) -> str:
        if not isinstance(self.content, str):
            return str(self.content)
        return self.content

    def render(self, **kwargs) -> "MessageType":
        """
        Renders the content as a jinja template with the given keyword arguments
        and returns a new Message.
        """
        content = jinja_env.from_string(self.content).render(**kwargs)
        return self.model_copy(update=dict(content=content))

    @classmethod
    def _langchain_message_kwargs(
        cls, message: langchain_core.messages.BaseMessage
    ) -> "BaseMessage":
        return message.dict(include={"content", "id", "name"}) | dict(
            _langchain_message=message
        )

    @classmethod
    def from_langchain_message(message: langchain_core.messages.BaseMessage, **kwargs):
        raise NotImplementedError()

    def to_langchain_message(self) -> langchain_core.messages.BaseMessage:
        raise NotImplementedError()


class AgentReference(ControlFlowModel):
    name: str


class AgentMessageMixin(ControlFlowModel):
    agent: Optional[AgentReference] = None

    @field_validator("agent", mode="before")
    def _validate_agent(cls, v):
        from controlflow.agents.agent import Agent

        if isinstance(v, Agent):
            return AgentReference(name=v.name)
        return v

    def __init__(self, *args, agent: "Agent" = None, **data):
        if agent is not None and data.get("name") is None:
            data["name"] = agent.name
        super().__init__(*args, agent=agent, **data)


class AIMessage(BaseMessage, AgentMessageMixin):
    role: Literal["ai"] = "ai"
    tool_calls: list[ToolCall] = []

    is_delta: bool = False

    def __init__(self, *args, **data):
        super().__init__(*args, **data)

        # GPT-4 models somtimes use a hallucinated parallel tool calling mechanism
        # whose name is not compatible with the API's restrictions on tool names
        for tool_call in self.tool_calls:
            if tool_call["name"] == "multi_tool_use.parallel":
                tool_call["name"] = "multi_tool_use_parallel"

    def has_tool_calls(self) -> bool:
        return any(self.tool_calls)

    @classmethod
    def from_langchain_message(
        cls,
        message: langchain_core.messages.AIMessage,
        **kwargs,
    ):
        data = dict(
            **cls._langchain_message_kwargs(message),
            tool_calls=message.tool_calls + getattr(message, "invalid_tool_calls", []),
        )

        return cls(**data | kwargs)

    def to_langchain_message(
        self,
    ) -> langchain_core.messages.AIMessage:
        if self._langchain_message is not None:
            return self._langchain_message
        return langchain_core.messages.AIMessage(
            content=self.content, tool_calls=self.tool_calls, id=self.id, name=self.name
        )


class AIMessageChunk(AIMessage):
    tool_calls: list[ToolCallChunk] = []

    @classmethod
    def from_langchain_message(
        cls,
        message: Union[
            langchain_core.messages.AIMessageChunk, langchain_core.messages.AIMessage
        ],
        **kwargs,
    ):
        if isinstance(message, langchain_core.messages.AIMessageChunk):
            tool_calls = message.tool_call_chunks
        elif isinstance(message, langchain_core.messages.AIMessage):
            tool_calls = []
            for i, call in enumerate(message.tool_calls):
                tool_calls.append(
                    ToolCallChunk(
                        id=call["id"],
                        name=call["name"],
                        args=str(call["args"]) if call["args"] else None,
                        index=call.get("index", i),
                    )
                )
        data = dict(
            **cls._langchain_message_kwargs(message),
            tool_calls=tool_calls,
            is_delta=True,
        )

        return cls(**data | kwargs)

    def to_langchain_message(
        self,
    ) -> langchain_core.messages.AIMessageChunk:
        if self._langchain_message is not None:
            return self._langchain_message
        return langchain_core.messages.AIMessageChunk(
            content=self.content,
            tool_call_chunks=self.tool_calls,
            id=self.id,
            name=self.name,
        )


class UserMessage(BaseMessage):
    role: Literal["user"] = "user"

    @classmethod
    def from_langchain_message(
        cls, message: langchain_core.messages.HumanMessage, **kwargs
    ):
        return cls(**cls._langchain_message_kwargs(message) | kwargs)

    def to_langchain_message(self) -> langchain_core.messages.BaseMessage:
        if self._langchain_message is not None:
            return self._langchain_message
        return langchain_core.messages.HumanMessage(
            content=self.content, id=self.id, name=self.name
        )


class SystemMessage(BaseMessage):
    role: Literal["system"] = "system"

    @classmethod
    def from_langchain_message(
        cls, message: langchain_core.messages.SystemMessage, **kwargs
    ):
        return cls(**cls._langchain_message_kwargs(message) | kwargs)

    def to_langchain_message(self) -> langchain_core.messages.BaseMessage:
        if self._langchain_message is not None:
            return self._langchain_message
        return langchain_core.messages.SystemMessage(
            content=self.content, id=self.id, name=self.name
        )


class ToolMessage(BaseMessage, AgentMessageMixin):
    model_config = dict(arbitrary_types_allowed=True)
    role: Literal["tool"] = "tool"
    tool_call_id: str
    tool_call: Union[ToolCall, InvalidToolCall]
    tool_result: Any = Field(None, exclude=True)
    tool_metadata: dict[str, Any] = {}
    is_error: bool = False

    def to_langchain_message(self) -> langchain_core.messages.BaseMessage:
        if self._langchain_message is not None:
            return self._langchain_message
        else:
            return langchain_core.messages.ToolMessage(
                id=self.id,
                name=self.name,
                content=self.content,
                tool_call_id=self.tool_call_id,
            )


MessageType = Union[UserMessage, AIMessage, SystemMessage, ToolMessage]



import inspect
import os
from datetime import datetime
from zoneinfo import ZoneInfo

from jinja2 import Environment as JinjaEnvironment
from jinja2 import StrictUndefined, select_autoescape

jinja_env = JinjaEnvironment(
    autoescape=select_autoescape(default_for_string=False),
    trim_blocks=True,
    lstrip_blocks=True,
    auto_reload=True,
    undefined=StrictUndefined,
)

jinja_env.globals.update(
    {
        "now": lambda: datetime.now(ZoneInfo("UTC")),
        "inspect": inspect,
        "getcwd": os.getcwd,
        "zip": zip,
    }
)



from controlflow import Agent, Task, flow

a1 = Agent(
    name="Half-full",
    instructions="""
    You are an eternal optimist.
    """,
)
a2 = Agent(
    name="Half-empty",
    instructions="""
    You are an eternal pessimist.
    """,
)
# create an agent that will decide who wins the debate
a3 = Agent(name="Moderator")


@flow
def demo():
    topic = "pineapple on pizza"

    task = Task(
        "Have a debate about the topic. Each agent should take at least two turns.",
        agents=[a1, a2],
        context={"topic": topic},
    )
    task.run()

    task2 = Task(
        "which argument do you find more compelling?",
        result_type=[a1.name, a2.name],
        agents=[a3],
    )
    task2.run()


if __name__ == "__main__":
    demo()



from enum import Enum
from typing import Any, Callable, Literal, Optional, TypeVar, Union

from pydantic import Field

from controlflow.agents import Agent
from controlflow.tasks.task import Task
from controlflow.utilities.types import ControlFlowModel

ToolLiteral = TypeVar("ToolLiteral", bound=str)


class ResultType(Enum):
    STRING = "STRING"
    NONE = "NONE"


class TaskReference(ControlFlowModel):
    """
    A reference to a task by its ID. Used for indicating task depenencies.
    """

    id: int


class AgentReference(ControlFlowModel):
    """
    A reference to an agent by its name. Used for assigning agents to tasks.
    """

    name: str


class AgentTemplate(ControlFlowModel):
    name: str
    description: Optional[str] = Field(
        None,
        description="A brief description of the agent that will be visible to other agents.",
    )
    instructions: Optional[str] = Field(
        None,
        description="Private instructions for the agent to follow when completing tasks.",
    )
    user_access: bool = Field(
        False, description="If True, the agent can interact with a human user."
    )
    tools: list[str] = Field([], description="The tools that the agent has access to.")


class TaskTemplate(ControlFlowModel):
    id: int
    objective: str = Field(description="The task's objective.")
    instructions: Optional[str] = Field(
        None, description="Instructions for completing the task."
    )
    result_type: Union[ResultType, list[str]] = Field(
        ResultType.STRING,
        description="The type of result expected from the task, defaults to a string output. "
        "Can also be `NONE` if the task does not produce a result (but may have side effects) or "
        "a list of choices if the task has a discrete set of possible outputs.",
    )
    context: dict[str, Union[TaskReference, Any]] = Field(
        default_factory=dict,
        description="The task's context. Values may be constants, TaskReferences, or "
        "collections of either. Any `TaskReferences` will create upstream dependencies, meaning "
        "this task will receive the referenced task's output as input.",
    )
    depends_on: list[TaskReference] = Field(
        default_factory=list,
        description="Tasks that must be completed before this task can be started, "
        "though their outputs are not used.",
    )
    parent: Optional[TaskReference] = Field(
        None,
        description="Indicate that this task is a subtask of a parent. Not required for top-level tasks.",
    )
    agents: list[AgentReference] = Field(
        default_factory=list,
        description="Any agents assigned to the task. If not specified, the default agent will be used.",
    )
    tools: list[str] = Field([], description="The tools available for this task.")
    user_access: bool = Field(
        False, description="If True, the task requires interaction with a human user."
    )


def create_tasks(
    task_templates: list[TaskTemplate],
    agent_templates: list[AgentTemplate] = None,
    agents: list[Agent] = None,
    tools: dict[str, Any] = None,
) -> list[Task]:
    """
    Create tasks from task templates, agent templates, agents, and tools.

    Task templates and agent templates are JSON-serializable objects that define the tasks and agents to be created.

    Agents and tools represent pre-existing agents and tools that can be used to complete the task and agent templates.
    """
    agents: dict[str, Agent] = {a.name: a for a in agents or []}
    tasks: dict[int, Task] = {}
    task_templates: dict[int, TaskTemplate] = {t.id: t for t in task_templates}

    # create agents from templates
    for agent_template in agent_templates:
        agents[agent_template.name] = Agent(
            name=agent_template.name,
            description=agent_template.description,
            instructions=agent_template.instructions,
            user_access=agent_template.user_access,
            tools=[tools[tool] for tool in agent_template.tools],
        )

    # create tasks from templates
    for task_template in task_templates.values():
        if task_template.result_type == ResultType.NONE:
            result_type = None
        elif result_type == ResultType.STRING:
            result_type = str
        else:
            result_type = task_template.result_type

        tasks[task_template.id] = Task(
            objective=task_template.objective,
            instructions=task_template.instructions,
            result_type=result_type,
            tools=[tools[tool] for tool in task_template.tools],
            use_access=task_template.user_access,
        )

    # resolve references
    for template_id, task in tasks.items():
        task_template = task_templates[template_id]
        if task_agents := [
            agents[agent_ref.name] for agent_ref in task_template.agents
        ]:
            task.agents = task_agents
        task.depends_on = [tasks[d.id] for d in task_template.depends_on]
        task.context = {
            key: tasks[value.id] if isinstance(value, TaskReference) else value
            for key, value in task_template.context.items()
        }

        if parent := tasks[task_template.parent.id] if task_template.parent else None:
            parent.add_subtask(task)

    return list(tasks.values())


class Templates(ControlFlowModel):
    task_templates: list[TaskTemplate]
    agent_templates: list[AgentTemplate]


def auto_tasks(
    description: str,
    available_agents: list[Agent] = None,
    available_tools: list[Callable] = None,
) -> list[Task]:
    tool_names = []
    for tool in available_tools or []:
        tool_names.append(tool.__name__)

    if tool_names:
        literal_tool_names = Literal[*tool_names]  # type: ignore
    else:
        literal_tool_names = None

    class TaskTemplate_Tools(TaskTemplate):
        tools: list[literal_tool_names] = Field(
            [], description="The tools available for this task."
        )

    class AgentTemplate_Tools(AgentTemplate):
        tools: list[literal_tool_names] = Field(
            [], description="The tools that the agent has access to."
        )

    class Templates_Tools(Templates):
        task_templates: list[TaskTemplate_Tools]
        agent_templates: list[AgentTemplate_Tools]

    task = Task(
        objective="""
        Generate the minimal set of tasks required to complete the provided
        `description` of an objective. Also reference any tools or agents (or
        create new agents) that your tasks require.
        """,
        instructions="""
        Each task will be executed by an agent like you, working in a workflow
        like this one. Your job is to define the workflow. Choose your tasks to
        be achievable by agents with the tools and skills you deem necessary.
        Create only as many tasks as you need.
        
        Each task should be well-defined, with a single objective and clear
        instructions. The tasks should be independent of each other, but may
        have dependencies on other tasks. If you do not choose
        agents for your tasks, the default agent will be used. Do not post messages, just return your
        result.
        """,
        result_type=Templates,
        context=dict(
            description=description,
            available_agents=available_agents,
            available_tools=available_tools,
        ),
    )

    task.run()

    return create_tasks(
        task_templates=task.result.task_templates,
        agent_templates=task.result.agent_templates,
        agents=available_agents,
        tools=available_tools,
    )



from contextlib import contextmanager
from typing import Union

import langchain_core.messages
from langchain_core.language_models.fake_chat_models import FakeMessagesListChatModel

import controlflow
from controlflow.flows.history import InMemoryHistory
from controlflow.llm.messages import BaseMessage, MessageType


class FakeLLM(FakeMessagesListChatModel):
    def set_responses(
        self, responses: list[Union[MessageType, langchain_core.messages.BaseMessage]]
    ):
        new_responses = []
        for msg in responses:
            if isinstance(msg, BaseMessage):
                msg = msg.to_langchain_message()
            new_responses.append(msg)
        self.responses = new_responses

    def bind_tools(self, *args, **kwargs):
        """When binding tools, passthrough"""
        return self

    def get_num_tokens_from_messages(self, messages: list) -> int:
        """Approximate token counter for messages"""
        return len(str(messages))


@contextmanager
def record_messages(
    remove_additional_kwargs: bool = True, remove_tool_call_chunks: bool = True
):
    """
    Context manager for recording all messages in a flow, useful for testing.


    with record_messages() as messages:
        cf.Task("say hello").run()

    assert messages[0].content == "Hello!"

    """
    history = InMemoryHistory(history={})
    old_default_history = controlflow.default_history
    controlflow.default_history = history

    messages = []

    try:
        yield messages
    finally:
        controlflow.default_history = old_default_history

        _messages_buffer = []
        for _, thread_messages in history.history.items():
            for message in thread_messages:
                message = message.copy()
                if hasattr(message, "additional_kwargs") and remove_additional_kwargs:
                    message.additional_kwargs = {}
                if hasattr(message, "tool_call_chunks") and remove_tool_call_chunks:
                    message.tool_call_chunks = []
                _messages_buffer.append(message)

        messages.extend(sorted(_messages_buffer, key=lambda m: m.timestamp))



import datetime
import uuid
from contextlib import contextmanager
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    GenericAlias,
    Literal,
    Optional,
    TypeVar,
    Union,
    _LiteralGenericAlias,
)

from prefect.context import TaskRunContext
from pydantic import (
    Field,
    PydanticSchemaGenerationError,
    TypeAdapter,
    field_serializer,
    field_validator,
    model_validator,
)

import controlflow
from controlflow.agents import Agent
from controlflow.instructions import get_instructions
from controlflow.tools import Tool
from controlflow.tools.talk_to_user import talk_to_user
from controlflow.utilities.context import ctx
from controlflow.utilities.logging import get_logger
from controlflow.utilities.prefect import PrefectTrackingTask
from controlflow.utilities.prefect import prefect_task as prefect_task
from controlflow.utilities.tasks import (
    collect_tasks,
    visit_task_collection,
)
from controlflow.utilities.types import (
    NOTSET,
    ControlFlowModel,
)

if TYPE_CHECKING:
    from controlflow.controllers.graph import Graph
    from controlflow.flows import Flow

T = TypeVar("T")
logger = get_logger(__name__)


def get_task_run_name() -> str:
    context = TaskRunContext.get()
    return f'Run {context.parameters["self"].friendly_name()}'


class TaskStatus(Enum):
    INCOMPLETE = "INCOMPLETE"
    SUCCESSFUL = "SUCCESSFUL"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"


class Task(ControlFlowModel):
    id: str = Field(default_factory=lambda: str(uuid.uuid4().hex[:5]))
    objective: str = Field(
        ..., description="A brief description of the required result."
    )
    instructions: Union[str, None] = Field(
        None, description="Detailed instructions for completing the task."
    )
    agents: Optional[list["Agent"]] = Field(
        None,
        description="The agents assigned to the task. If not provided, agents "
        "will be inferred from the parent task, flow, or global default.",
        validate_default=True,
    )
    context: dict = Field(
        default_factory=dict,
        description="Additional context for the task. If tasks are provided as "
        "context, they are automatically added as `depends_on`",
    )
    parent: Optional["Task"] = Field(
        None,
        description="The parent task of this task. Subtasks are considered"
        " upstream dependencies of their parents.",
        validate_default=True,
    )
    depends_on: set["Task"] = Field(
        default_factory=set, description="Tasks that this task depends on explicitly."
    )
    status: TaskStatus = TaskStatus.INCOMPLETE
    result: T = None
    result_type: Union[type[T], GenericAlias, _LiteralGenericAlias, None] = Field(
        str,
        description="The expected type of the result. This should be a type"
        ", generic alias, BaseModel subclass, pd.DataFrame, or pd.Series. "
        "Can be None if no result is expected or the agent should communicate internally.",
    )
    error: Union[str, None] = None
    tools: list[Callable] = Field(
        default_factory=list,
        description="Tools available to every agent working on this task.",
    )
    user_access: bool = False
    private: bool = False
    agent_strategy: Optional[Callable] = Field(
        None,
        description="A function that returns an agent, used for customizing how "
        "the next agent is selected. The returned agent must be one "
        "of the assigned agents. If not provided, will be inferred "
        "from the parent task; round-robin selection is the default. "
        "Only used for tasks with more than one agent assigned.",
    )
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)
    _subtasks: set["Task"] = set()
    _downstreams: set["Task"] = set()
    _iteration: int = 0
    _cm_stack: list[contextmanager] = []
    _prefect_task: Optional[PrefectTrackingTask] = None

    model_config = dict(extra="forbid", arbitrary_types_allowed=True)

    def __init__(
        self,
        objective: str = None,
        result_type: Any = NOTSET,
        infer_parent: bool = True,
        **kwargs,
    ):
        # allow certain args to be provided as a positional args
        if result_type is not NOTSET:
            kwargs["result_type"] = result_type
        if objective is not None:
            kwargs["objective"] = objective
        # if parent is None and infer parent is False, set parent to NOTSET
        if not infer_parent and kwargs.get("parent") is None:
            kwargs["parent"] = NOTSET
        if additional_instructions := get_instructions():
            kwargs["instructions"] = (
                kwargs.get("instructions")
                or "" + "\n" + "\n".join(additional_instructions)
            ).strip()

        super().__init__(**kwargs)

        self._prefect_task = PrefectTrackingTask(
            name=f"Working on {self.friendly_name()}...",
            description=self.instructions,
            tags=[self.__class__.__name__],
        )

    def __hash__(self):
        return hash((self.__class__.__name__, self.id))

    def __eq__(self, other):
        """
        Tasks have set attributes and set equality is based on id() of their
        contents, not equality of objects. This means that two tasks are not
        equal unless their set attributes satisfy an identity criteria, which is
        too strict.
        """
        if type(self) == type(other):
            d1 = dict(self)
            d2 = dict(other)
            # conver sets to lists for comparison
            d1["depends_on"] = list(d1["depends_on"])
            d2["depends_on"] = list(d2["depends_on"])
            return d1 == d2
        return False

    def __repr__(self) -> str:
        serialized = self.model_dump()
        return f"{self.__class__.__name__}({', '.join(f'{key}={repr(value)}' for key, value in serialized.items())})"

    @field_validator("parent", mode="before")
    def _default_parent(cls, v):
        if v is None:
            parent_tasks = ctx.get("tasks", [])
            v = parent_tasks[-1] if parent_tasks else None
        elif v is NOTSET:
            v = None
        return v

    @field_validator("agents", mode="before")
    def _validate_agents(cls, v):
        if v == []:
            raise ValueError("At least one agent is required.")
        return v

    @field_validator("result_type", mode="before")
    def _turn_list_into_literal_result_type(cls, v):
        if isinstance(v, (list, tuple, set)):
            return Literal[tuple(v)]  # type: ignore
        return v

    @model_validator(mode="after")
    def _finalize(self):
        # add task to flow, if exists
        if flow := controlflow.flows.get_flow():
            flow.add_task(self)

        # create dependencies to tasks passed in as depends_on
        for task in self.depends_on:
            self.add_dependency(task)

        # create dependencies to tasks passed as subtasks
        if self.parent is not None:
            self.parent.add_subtask(self)

        # create dependencies to tasks passed in as context
        context_tasks = collect_tasks(self.context)

        for task in context_tasks:
            self.depends_on.add(task)

        return self

    @field_serializer("parent")
    def _serialize_parent(self, parent: Optional["Task"]):
        return parent.id if parent is not None else None

    @field_serializer("depends_on")
    def _serialize_depends_on(self, depends_on: set["Task"]):
        return [t.id for t in depends_on]

    @field_serializer("context")
    def _serialize_context(self, context: dict):
        def visitor(task):
            return f"<Result from task {task.id}>"

        return visit_task_collection(context, visitor)

    @field_serializer("result_type")
    def _serialize_result_type(self, result_type: list["Task"]):
        if result_type is None:
            return None
        try:
            schema = TypeAdapter(result_type).json_schema()
        except PydanticSchemaGenerationError:
            schema = "<schema could not be generated>"

        return dict(type=repr(result_type), schema=schema)

    @field_serializer("agents")
    def _serialize_agents(self, agents: Optional[list["Agent"]]):
        agents = self.get_agents()
        return [a.serialize_for_prompt() for a in agents]

    @field_serializer("tools")
    def _serialize_tools(self, tools: list[Callable]):
        tools = controlflow.tools.as_tools(tools)
        # tools are Pydantic 1 objects
        return [t.dict(include={"name", "description"}) for t in tools]

    def friendly_name(self):
        if len(self.objective) > 50:
            objective = f'"{self.objective[:50]}..."'
        else:
            objective = f'"{self.objective}"'
        return f"Task {self.id} ({objective})"

    def as_graph(self) -> "Graph":
        from controlflow.controllers.graph import Graph

        return Graph.from_tasks(tasks=[self])

    @property
    def subtasks(self) -> list["Task"]:
        from controlflow.controllers.graph import Graph

        return Graph.from_tasks(tasks=self._subtasks).topological_sort()

    def add_subtask(self, task: "Task"):
        """
        Indicate that this task has a subtask (which becomes an implicit dependency).
        """
        if task.parent is None:
            task.parent = self
        elif task.parent is not self:
            raise ValueError(f"{self.friendly_name()} already has a parent.")
        self._subtasks.add(task)
        self.depends_on.add(task)

    def add_dependency(self, task: "Task"):
        """
        Indicate that this task depends on another task.
        """
        self.depends_on.add(task)
        task._downstreams.add(self)

    def run_once(self, agents: Optional[list["Agent"]] = None, flow: "Flow" = None):
        """
        Runs the task with provided agent. If no agent is provided, one will be selected from the task's agents.
        """
        # run once doesn't create new flows because the history would be lost
        flow = flow or controlflow.flows.get_flow()
        if flow is None:
            raise ValueError(
                "Task.run_once() must be called within a flow context or with a flow argument."
            )

        from controlflow.controllers import Controller

        controller = Controller(
            tasks=[self], flow=flow, agents={self: agents} if agents else None
        )
        controller.run_once()

    async def run_once_async(
        self, agents: Optional[list["Agent"]] = None, flow: "Flow" = None
    ):
        """
        Runs the task with provided agent. If no agent is provided, one will be selected from the task's agents.
        """

        # run once doesn't create new flows because the history would be lost
        flow = flow or controlflow.flows.get_flow()
        if flow is None:
            raise ValueError(
                "Task.run_once_async() must be called within a flow context or with a flow argument."
            )

        from controlflow.controllers import Controller

        controller = Controller(
            tasks=[self], flow=flow, agents={self: agents} if agents else None
        )
        await controller.run_once_async()

    @prefect_task(task_run_name=get_task_run_name)
    def run(
        self,
        agents: Optional[list["Agent"]] = None,
        raise_on_error: bool = True,
        flow: "Flow" = None,
    ) -> Generator[T, None, None]:
        """
        Internal function that can handle both sync and async runs by yielding either the result or the coroutine.
        """
        from controlflow.flows import Flow, get_flow

        flow = flow or get_flow()
        if flow is None:
            if controlflow.settings.strict_flow_context:
                raise ValueError(
                    "Task.run() must be called within a flow context or with a "
                    "flow argument if implicit flows are disabled."
                )
            else:
                flow = Flow()

        from controlflow.controllers import Controller

        controller = Controller(
            tasks=[self], flow=flow, agents={self: agents} if agents else None
        )
        controller.run()

        if self.is_successful():
            return self.result
        elif self.is_failed() and raise_on_error:
            raise ValueError(f"{self.friendly_name()} failed: {self.error}")

    @prefect_task(task_run_name=get_task_run_name)
    async def run_async(
        self,
        agents: Optional[list["Agent"]] = None,
        raise_on_error: bool = True,
        flow: "Flow" = None,
    ) -> Generator[T, None, None]:
        """
        Internal function that can handle both sync and async runs by yielding either the result or the coroutine.
        """
        from controlflow.flows import Flow, get_flow

        flow = flow or get_flow()
        if flow is None:
            if controlflow.settings.strict_flow_context:
                raise ValueError(
                    "Task.run() must be called within a flow context or with a "
                    "flow argument if implicit flows are disabled."
                )
            else:
                flow = Flow()

        from controlflow.controllers import Controller

        controller = Controller(
            tasks=[self], flow=flow, agents={self: agents} if agents else None
        )
        await controller.run_async()

        if self.is_successful():
            return self.result
        elif self.is_failed() and raise_on_error:
            raise ValueError(f"{self.friendly_name()} failed: {self.error}")

    def copy(self):
        return self.model_copy()

    @contextmanager
    def create_context(self):
        stack = ctx.get("tasks", [])
        with ctx(tasks=stack + [self]):
            yield self

    def __enter__(self):
        # use stack so we can enter the context multiple times
        self._cm_stack.append(self.create_context())
        return self._cm_stack[-1].__enter__()

    def __exit__(self, *exc_info):
        return self._cm_stack.pop().__exit__(*exc_info)

    def is_incomplete(self) -> bool:
        return self.status == TaskStatus.INCOMPLETE

    def is_complete(self) -> bool:
        return self.status != TaskStatus.INCOMPLETE

    def is_successful(self) -> bool:
        return self.status == TaskStatus.SUCCESSFUL

    def is_failed(self) -> bool:
        return self.status == TaskStatus.FAILED

    def is_skipped(self) -> bool:
        return self.status == TaskStatus.SKIPPED

    def is_ready(self) -> bool:
        """
        Returns True if all dependencies are complete and this task is incomplete.
        """
        return self.is_incomplete() and all(t.is_complete() for t in self.depends_on)

    def _create_success_tool(self) -> Tool:
        """
        Create an agent-compatible tool for marking this task as successful.
        """
        # generate tool for result_type=None
        if self.result_type is None:

            def succeed() -> str:
                return self.mark_successful(result=None)

        # generate tool for other result types
        else:
            result_schema = generate_result_schema(self.result_type)

            def succeed(result: result_schema) -> str:  # type: ignore
                return self.mark_successful(result=result)

        return Tool.from_function(
            succeed,
            name=f"mark_task_{self.id}_successful",
            description=f"Mark task {self.id} as successful.",
            metadata=dict(ignore_result=True),
        )

    def _create_fail_tool(self) -> Tool:
        """
        Create an agent-compatible tool for failing this task.
        """

        return Tool.from_function(
            self.mark_failed,
            name=f"mark_task_{self.id}_failed",
            description=f"Mark task {self.id} as failed. Only use when technical errors prevent success.",
            metadata=dict(ignore_result=True),
        )

    def _create_skip_tool(self) -> Tool:
        """
        Create an agent-compatible tool for skipping this task.
        """
        return Tool.from_function(
            self.mark_skipped,
            name=f"mark_task_{self.id}_skipped",
            description=f"Mark task {self.id} as skipped. Only use when completing a parent task early.",
            metadata=dict(ignore_result=True),
        )

    def get_agents(self) -> list["Agent"]:
        if self.agents:
            return self.agents
        elif self.parent:
            return self.parent.get_agents()
        else:
            from controlflow.agents import get_default_agent
            from controlflow.flows import get_flow

            try:
                flow = get_flow()
            except ValueError:
                flow = None
            if flow and flow.agents:
                return flow.agents
            else:
                return [get_default_agent()]

    def get_agent_strategy(self) -> Callable:
        """
        Get a function for selecting the next agent to work on this
        task.

        If an agent_strategy is provided, it will be used. Otherwise, the parent
        task's agent_strategy will be used. Finally, the global default agent_strategy
        will be used (round-robin selection).
        """
        if self.agent_strategy is not None:
            return self.agent_strategy
        elif self.parent:
            return self.parent.get_agent_strategy()
        else:
            import controlflow.tasks.agent_strategies

            return controlflow.tasks.agent_strategies.round_robin

    def get_tools(self) -> list[Union[Tool, Callable]]:
        tools = self.tools.copy()
        # if this task is ready to run, generate tools
        if self.is_ready():
            tools.extend([self._create_fail_tool(), self._create_success_tool()])
            # add skip tool if this task has a parent task
            # if self.parent is not None:
            #     tools.append(self._create_skip_tool())
        if self.user_access:
            tools.append(talk_to_user)
        return tools
        # return [wrap_prefect_tool(t) for t in tools]

    def set_status(self, status: TaskStatus):
        self.status = status

        # update TUI
        if tui := ctx.get("tui"):
            tui.update_task(self)

        # update Prefect
        if self._prefect_task.is_started:
            if status == TaskStatus.SUCCESSFUL:
                self._prefect_task.succeed(self.result)
            elif status == TaskStatus.FAILED:
                self._prefect_task.fail(self.error)
            elif status == TaskStatus.SKIPPED:
                self._prefect_task.skip()

    def mark_successful(self, result: T = None, validate_upstreams: bool = True):
        if validate_upstreams:
            if any(t.is_incomplete() for t in self.depends_on):
                raise ValueError(
                    f"Task {self.objective} cannot be marked successful until all of its "
                    "upstream dependencies are completed. Incomplete dependencies "
                    f"are: {', '.join(t.friendly_name() for t in self.depends_on if t.is_incomplete())}"
                )
            elif any(t.is_incomplete() for t in self._subtasks):
                raise ValueError(
                    f"Task {self.objective} cannot be marked successful until all of its "
                    "subtasks are completed. Incomplete subtasks "
                    f"are: {', '.join(t.friendly_name() for t in self._subtasks if t.is_incomplete())}"
                )

        self.result = validate_result(result, self.result_type)
        self.set_status(TaskStatus.SUCCESSFUL)

        if agent := ctx.get("agent"):
            return f"{self.friendly_name()} marked successful by {agent.name}."
        return f"{self.friendly_name()} marked successful."

    def mark_failed(self, message: Union[str, None] = None):
        self.error = message
        self.set_status(TaskStatus.FAILED)
        if agent := ctx.get("agent"):
            return f"{self.friendly_name()} marked failed by {agent.name}."
        return f"{self.friendly_name()} marked failed."

    def mark_skipped(self):
        self.set_status(TaskStatus.SKIPPED)
        if agent := ctx.get("agent"):
            return f"{self.friendly_name()} marked skipped by {agent.name}."
        return f"{self.friendly_name()} marked skipped."

    def generate_subtasks(self, instructions: str = None, agent: Agent = None):
        """
        Generate subtasks for this task based on the provided instructions.
        Subtasks can reuse the same tools and agents as this task.
        """
        from controlflow.planning.plan import create_plan

        # enter a context to set the parent task
        with self:
            create_plan(
                self.objective,
                instructions=instructions,
                planning_agent=agent or self.agents,
                agents=self.agents,
                tools=self.tools,
                context=self.context,
            )


def generate_result_schema(result_type: type[T]) -> type[T]:
    result_schema = None
    # try loading pydantic-compatible schemas
    try:
        TypeAdapter(result_type)
        result_schema = result_type
    except PydanticSchemaGenerationError:
        pass
    # try loading as dataframe
    # try:
    #     import pandas as pd

    #     if result_type is pd.DataFrame:
    #         result_schema = PandasDataFrame
    #     elif result_type is pd.Series:
    #         result_schema = PandasSeries
    # except ImportError:
    #     pass
    if result_schema is None:
        raise ValueError(
            f"Could not load or infer schema for result type {result_type}. "
            "Please use a custom type or add compatibility."
        )
    return result_schema


def validate_result(result: Any, result_type: type[T]) -> T:
    if result_type is None and result is not None:
        raise ValueError("Task has result_type=None, but a result was provided.")
    elif result_type is not None:
        try:
            result = TypeAdapter(result_type).validate_python(result)
        except PydanticSchemaGenerationError:
            if isinstance(result, dict):
                result = result_type(**result)
            else:
                result = result_type(result)

        # Convert DataFrame schema back into pd.DataFrame object
        # if result_type == PandasDataFrame:
        #     import pandas as pd

        #     result = pd.DataFrame(**result)
        # elif result_type == PandasSeries:
        #     import pandas as pd

        #     result = pd.Series(**result)

    return result



import controlflow
import pytest
from controlflow.llm.messages import MessageType
from controlflow.utilities.testing import FakeLLM


@pytest.fixture(autouse=True)
def restore_defaults(monkeypatch):
    """
    Monkeypatch defaults to themselves, which will automatically reset them after every test
    """
    monkeypatch.setattr(controlflow, "default_agent", controlflow.default_agent)
    monkeypatch.setattr(controlflow, "default_model", controlflow.default_model)
    monkeypatch.setattr(controlflow, "default_history", controlflow.default_history)
    yield


@pytest.fixture()
def fake_llm() -> FakeLLM:
    return FakeLLM(responses=[])


@pytest.fixture()
def default_fake_llm(fake_llm) -> FakeLLM:
    controlflow.default_model = fake_llm
    return fake_llm



from typing import Optional, Union

import prefect
from pydantic import BaseModel, ConfigDict

# flag for unset defaults
NOTSET = "__NOTSET__"


class ControlFlowModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        extra="forbid",
        ignored_types=(prefect.Flow, prefect.Task),
    )


class PandasDataFrame(ControlFlowModel):
    """Schema for a pandas dataframe"""

    data: Union[
        list[list[Union[str, int, float, bool]]],
        dict[str, list[Union[str, int, float, bool]]],
    ]
    columns: Optional[list[str]] = None
    index: Optional[list[str]] = None
    dtype: Optional[dict[str, str]] = None


class PandasSeries(ControlFlowModel):
    """Schema for a pandas series"""

    data: list[Union[str, int, float]]
    index: Optional[list[str]] = None
    name: Optional[str] = None
    dtype: Optional[str] = None


class _OpenAIBaseType(ControlFlowModel):
    model_config = ConfigDict(extra="allow")


__all__ = ["ControlFlowModel", "PandasDataFrame", "PandasSeries", "_OpenAIBaseType"]



import functools
import inspect
import json
import typing
from typing import TYPE_CHECKING, Annotated, Any, Callable, Optional, Union

import langchain_core.tools
import pydantic
import pydantic.v1
from langchain_core.messages import ToolCall
from prefect.utilities.asyncutils import run_coro_as_sync
from pydantic import Field, TypeAdapter

import controlflow
from controlflow.llm.messages import ToolMessage
from controlflow.utilities.prefect import create_markdown_artifact, prefect_task
from controlflow.utilities.types import ControlFlowModel

if TYPE_CHECKING:
    from controlflow.agents import Agent


TOOL_CALL_FUNCTION_RESULT_TEMPLATE = """
# Tool call: {name}

**Description:** {description}

## Arguments

```json
{args}
```

## Result

```
{result}
```
"""


class Tool(ControlFlowModel):
    name: str
    description: str
    parameters: dict
    metadata: dict = Field({}, exclude_none=True)
    fn: Callable = Field(None, exclude=True)

    def to_lc_tool(self) -> dict:
        return self.model_dump(include={"name", "description", "parameters"})

    @prefect_task(task_run_name="Tool call: {self.name}")
    def run(self, input: dict):
        result = self.fn(**input)
        if inspect.isawaitable(result):
            result = run_coro_as_sync(result)

        # prepare artifact
        passed_args = inspect.signature(self.fn).bind(**input).arguments
        try:
            # try to pretty print the args
            passed_args = json.dumps(passed_args, indent=2)
        except Exception:
            pass
        create_markdown_artifact(
            markdown=TOOL_CALL_FUNCTION_RESULT_TEMPLATE.format(
                name=self.name,
                description=self.description or "(none provided)",
                args=passed_args,
                result=result,
            ),
            key="tool-result",
        )
        return result

    @classmethod
    def from_function(
        cls, fn: Callable, name: str = None, description: str = None, **kwargs
    ):
        description = description or fn.__doc__ or "(No description provided)"

        signature = inspect.signature(fn)
        parameters = TypeAdapter(fn).json_schema()

        # load parameter descriptions
        for param in signature.parameters.values():
            # handle Annotated type hints
            if typing.get_origin(param.annotation) is Annotated:
                param_description = " ".join(
                    str(a) for a in typing.get_args(param.annotation)[1:]
                )
            # handle pydantic Field descriptions
            elif param.default is not inspect.Parameter.empty and isinstance(
                param.default, pydantic.fields.FieldInfo
            ):
                param_description = param.default.description
            else:
                param_description = None

            if param_description:
                parameters["properties"][param.name]["description"] = param_description

        # Handle return type description
        return_type = signature.return_annotation
        if return_type is not inspect._empty:
            return_schema = TypeAdapter(return_type).json_schema()
            description += f"\n\nReturn value schema: {return_schema}"

        return cls(
            name=name or fn.__name__,
            description=description,
            parameters=parameters,
            fn=fn,
            **kwargs,
        )

    @classmethod
    def from_lc_tool(cls, tool: langchain_core.tools.BaseTool, **kwargs):
        if isinstance(tool, langchain_core.tools.StructuredTool):
            fn = tool.func
        else:
            fn = lambda *a, **k: None  # noqa
        return cls(
            name=tool.name,
            description=tool.description,
            parameters=tool.args_schema.schema(),
            fn=fn,
            **kwargs,
        )


def tool(
    fn: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    metadata: Optional[dict] = None,
) -> Tool:
    """
    Decorator for turning a function into a Tool
    """
    if fn is None:
        return functools.partial(
            tool, name=name, description=description, metadata=metadata
        )
    return Tool.from_function(
        fn, name=name, description=description, metadata=metadata or {}
    )


def as_tools(
    tools: list[Union[Callable, langchain_core.tools.BaseTool, Tool]],
) -> list[Tool]:
    """
    Converts a list of tools (either Tool objects or callables) into a list of
    Tool objects.

    If duplicate tools are found, where the name, function, and coroutine are
    the same, only one is kept.
    """
    seen = set()
    new_tools = []
    for t in tools:
        if isinstance(t, Tool):
            pass
        elif isinstance(t, langchain_core.tools.BaseTool):
            t = Tool.from_lc_tool(t)
        elif inspect.isfunction(t):
            t = Tool.from_function(t)
        elif isinstance(t, dict):
            t = Tool(**t)
        else:
            raise ValueError(f"Invalid tool: {t}")

        if (t.name, t.description, t.fn) in seen:
            continue
        new_tools.append(t)
        seen.add((t.name, t.description, t.fn))
    return new_tools


def output_to_string(output: Any) -> str:
    """
    Function outputs must be provided as strings
    """
    if output is None:
        return ""
    elif isinstance(output, str):
        return output
    try:
        return pydantic.TypeAdapter(type(output)).dump_json(output).decode()
    except Exception:
        return str(output)


def handle_tool_call(
    tool_call: ToolCall,
    tools: list[Tool],
    error: str = None,
    agent: "Agent" = None,
) -> ToolMessage:
    tool_lookup = {t.name: t for t in tools}
    fn_name = tool_call["name"]
    is_error = False
    metadata = {}
    try:
        if error:
            fn_output = error
            is_error = True
        elif fn_name not in tool_lookup:
            fn_output = f'Function "{fn_name}" not found.'
            is_error = True
        else:
            tool = tool_lookup[fn_name]
            metadata.update(getattr(tool, "metadata", {}))
            fn_args = tool_call["args"]
            if isinstance(tool, Tool):
                fn_output = tool.run(input=fn_args)
            elif isinstance(tool, langchain_core.tools.BaseTool):
                fn_output = tool.invoke(input=fn_args)
    except Exception as exc:
        fn_output = f'Error calling function "{fn_name}": {exc}'
        is_error = True
        if controlflow.settings.tools_raise_on_error:
            raise

    from controlflow.llm.messages import ToolMessage

    return ToolMessage(
        content=output_to_string(fn_output),
        tool_call_id=tool_call["id"],
        tool_call=tool_call,
        tool_result=fn_output,
        tool_metadata=metadata,
        is_error=is_error,
        agent=agent,
    )



import functools
import inspect
from typing import Any, Callable, Optional, Union

import controlflow
from controlflow.agents import Agent
from controlflow.flows import Flow
from controlflow.tasks.task import Task
from controlflow.utilities.logging import get_logger
from controlflow.utilities.prefect import prefect_flow as prefect_flow
from controlflow.utilities.prefect import prefect_task as prefect_task

# from controlflow.utilities.marvin import patch_marvin
from controlflow.utilities.tasks import resolve_tasks

logger = get_logger(__name__)


def flow(
    fn: Optional[Callable[..., Any]] = None,
    *,
    thread: Optional[str] = None,
    instructions: Optional[str] = None,
    tools: Optional[list[Callable[..., Any]]] = None,
    agents: Optional[list[Agent]] = None,
    lazy: Optional[bool] = None,
    retries: Optional[int] = None,
    retry_delay_seconds: Optional[Union[float, int]] = None,
    timeout_seconds: Optional[Union[float, int]] = None,
    kwargs: Optional[dict[str, Any]] = None,
):
    """
    A decorator that wraps a function as a ControlFlow flow.

    When the function is called, a new flow is created and any tasks created
    within the function will be run as part of that flow. When the function
    returns, all tasks created in the flow will be run to completion (if they
    were not already completed) and their results will be returned. Any tasks
    that are returned from the function will be replaced with their resolved
    result.

    Args:
        fn (callable, optional): The function to be wrapped as a flow. If not provided,
            the decorator will act as a partial function and return a new flow decorator.
        thread (str, optional): The thread to execute the flow on. Defaults to None.
        instructions (str, optional): Instructions for the flow. Defaults to None.
        tools (list[Callable], optional): List of tools to be used in the flow. Defaults to None.
        agents (list[Agent], optional): List of agents to be used in the flow. Defaults to None.
        lazy (bool, optional): Whether the flow should be run lazily. If not
            set, behavior is determined by the global `eager_mode` setting. Lazy execution means
            that tasks are not run and a `Flow` object is returned instead.

    Returns:
        callable: The wrapped function or a new flow decorator if `fn` is not provided.
    """
    ...

    if fn is None:
        return functools.partial(
            flow,
            thread=thread,
            instructions=instructions,
            tools=tools,
            agents=agents,
            lazy=lazy,
            retries=retries,
            retry_delay_seconds=retry_delay_seconds,
            timeout_seconds=timeout_seconds,
            kwargs=kwargs,
        )

    sig = inspect.signature(fn)

    # the flow decorator creates a proper prefect flow
    @prefect_flow(
        timeout_seconds=timeout_seconds,
        retries=retries,
        retry_delay_seconds=retry_delay_seconds,
        **kwargs or {},
    )
    @functools.wraps(fn)
    def wrapper(
        *args,
        flow_kwargs: dict = None,
        lazy_: bool = False,
        **kwargs,
    ):
        # first process callargs
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        flow_kwargs = flow_kwargs or {}

        if thread is not None:
            flow_kwargs.setdefault("thread", thread)
        if tools is not None:
            flow_kwargs.setdefault("tools", tools)
        if agents is not None:
            flow_kwargs.setdefault("agents", agents)

        flow_obj = Flow(
            name=fn.__name__,
            description=fn.__doc__,
            context=bound.arguments,
            **flow_kwargs,
        )

        with flow_obj.create_context(create_prefect_flow_context=False):
            with controlflow.instructions(instructions):
                result = fn(*args, **kwargs)

                # Determine if we should run eagerly or lazily
                if lazy_ is not None:
                    run_eagerly = not lazy_
                elif lazy is not None:
                    run_eagerly = not lazy
                else:
                    run_eagerly = controlflow.settings.eager_mode

                if run_eagerly:
                    flow_obj.run()

                    # resolve any returned tasks; this will raise on failure
                    return resolve_tasks(result)
                else:
                    return flow_obj

    if lazy is True or (lazy is None and not controlflow.settings.eager_mode):
        wrapper.__annotations__["return"] = Flow

    return wrapper


def task(
    fn: Optional[Callable[..., Any]] = None,
    *,
    objective: Optional[str] = None,
    instructions: Optional[str] = None,
    agents: Optional[list["Agent"]] = None,
    tools: Optional[list[Callable[..., Any]]] = None,
    user_access: Optional[bool] = None,
    lazy: Optional[bool] = None,
    retries: Optional[int] = None,
    retry_delay_seconds: Optional[Union[float, int]] = None,
    timeout_seconds: Optional[Union[float, int]] = None,
    **task_kwargs: Optional[dict[str, Any]],
):
    """
    A decorator that turns a Python function into a Task. The Task objective is
    set to the function name, and the instructions are set to the function
    docstring. When the function is called in eager mode (default), the arguments are
    provided to the task as context, and the task is run to completion. If
    successful, the task result is returned; if failed, an error is raised. When
    the function is called with eager mode disabled or `lazy=True`, a Task object is
    returned which can be run later.

    Args:
        fn (callable, optional): The function to be wrapped as a task. If not provided,
            the decorator will act as a partial function and return a new task decorator.
        objective (str, optional): The objective of the task. Defaults to None, in which
            case the function name is used as the objective.
        instructions (str, optional): Instructions for the task. Defaults to None, in which
            case the function docstring is used as the instructions.
        agents (list[Agent], optional): List of agents to be used in the task. Defaults to None.
        tools (list[Callable], optional): List of tools to be used in the task. Defaults to None.
        user_access (bool, optional): Whether the task requires user access. Defaults to None,
            in which case it is set to False.
        lazy (bool, optional): Whether the task should be run lazily. If not
            set, behavior is determined by the global `eager_mode` setting. Lazy
            execution means that a `Task` object is returned instead of running the task.

    Returns:
        callable: The wrapped function or a new task decorator if `fn` is not provided.
    """

    if fn is None:
        return functools.partial(
            task,
            objective=objective,
            instructions=instructions,
            agents=agents,
            tools=tools,
            user_access=user_access,
            lazy=lazy,
            retries=retries,
            retry_delay_seconds=retry_delay_seconds,
            timeout_seconds=timeout_seconds,
            **task_kwargs,
        )

    sig = inspect.signature(fn)

    if objective is None:
        objective = fn.__name__

    if instructions is None:
        instructions = fn.__doc__

    result_type = fn.__annotations__.get("return")

    @prefect_task(
        timeout_seconds=timeout_seconds,
        retries=retries,
        retry_delay_seconds=retry_delay_seconds,
    )
    @functools.wraps(fn)
    def wrapper(*args, lazy_: bool = None, **kwargs):
        # first process callargs
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        task = Task(
            objective=objective,
            instructions=instructions,
            agents=agents,
            context=bound.arguments,
            result_type=result_type,
            user_access=user_access or False,
            tools=tools or [],
            **task_kwargs,
        )

        # Determine if we should run eagerly or lazily
        if lazy_ is not None:
            run_eagerly = not lazy_
        elif lazy is not None:
            run_eagerly = not lazy
        else:
            run_eagerly = controlflow.settings.eager_mode

        if run_eagerly:
            task.run()
            return task.result
        else:
            return task

    if lazy is True or (lazy is None and not controlflow.settings.eager_mode):
        wrapper.__annotations__["return"] = Task

    return wrapper
